# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Module defining a Digital RF Source block."""
import os
import sys
import warnings
from collections import defaultdict
from distutils.version import LooseVersion
from itertools import chain, tee

import numpy as np
import pmt
import six
from gnuradio import gr
from six.moves import zip

from digital_rf import (DigitalMetadataWriter, DigitalRFWriter,
                        _py_rf_write_hdf5, util)


def parse_time_pmt(val, samples_per_second):
    """Get (sec, frac, idx) from an rx_time pmt value."""
    tsec = np.uint64(pmt.to_uint64(pmt.tuple_ref(val, 0)))
    tfrac = pmt.to_double(pmt.tuple_ref(val, 1))
    # calculate sample index of time and floor to uint64
    tidx = np.uint64(tsec*samples_per_second + tfrac*samples_per_second)
    return int(tsec), tfrac, int(tidx)


def translate_rx_freq(tag):
    """Translate 'rx_freq' tag to 'center_frequencies' metadata sample."""
    offset = tag.offset
    key = 'center_frequencies'
    val = np.array(pmt.to_python(tag.value), ndmin=1)
    yield offset, key, val


def translate_metadata(tag):
    """Translate 'metadata' dictionary tag to metadata samples."""
    offset = tag.offset
    md = pmt.to_python(tag.value)
    try:
        for key, val in md.items():
            yield offset, key, val
    except AttributeError:
        wrnstr = (
            "Received 'metadata' stream tag that isn't a dictionary. Ignoring."
        )
        warnings.warn(wrnstr)


def collect_tags_in_dict(tags, translator, tag_dict={}):
    """Add the stream tags to `tag_dict` by their offset."""
    for tag in tags:
        for offset, key, val in translator(tag):
            # add tag as its own dictionary to tag_dict[offset]
            tag_dict.setdefault(offset, {}).update(((key, val),))


def recursive_dict_update(d, u):
    """Update d with values from u, recursing into sub-dictionaries."""
    for k, v in u.items():
        if isinstance(v, dict):
            try:
                # copy because we don't want to modify the sub-dictionary
                # just use its values to create an updated sub-dictionary
                subdict = d[k].copy()
            except KeyError:
                subdict = {}
            d[k] = recursive_dict_update(subdict, v)
        else:
            d[k] = v
    return d


def pairwise(iterable):
    """Return iterable elements in pairs, e.g. range(3) -> (0, 1), (1, 2)."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class digital_rf_channel_sink(gr.sync_block):
    """Sink block for writing a channel of Digital RF data."""

    def __init__(
        self, channel_dir, dtype, subdir_cadence_secs,
        file_cadence_millisecs, sample_rate_numerator, sample_rate_denominator,
        start=None, ignore_tags=False, is_complex=True, num_subchannels=1,
        uuid_str=None, center_frequencies=None, metadata={},
        is_continuous=True, compression_level=0,
        checksum=False, marching_periods=True, stop_on_skipped=True,
        debug=False, min_chunksize=None,
    ):
        """Write a channel of data in Digital RF format.

        In addition to storing the input samples in Digital RF format, this
        block also populates the channel's accompanying Digital Metadata
        at the sample indices when the metadata changes or a data skip occurs.
        See the Notes section for details on what metadata is stored.

        Parameters
        ----------

        channel_dir : string
            The directory where this channel is to be written. It will be
            created if it does not exist. The basename (last component) of the
            path is considered the channel's name for reading purposes.

        dtype : numpy.dtype | object to be cast by numpy.dtype()
            Object that gives the numpy dtype of the data to be written. This
            value is passed into ``numpy.dtype`` to get the actual dtype
            (e.g. ``numpy.dtype('>i4')``). Scalar types, complex types, and
            structured complex types with 'r' and 'i' fields of scalar types
            are valid.

        subdir_cadence_secs : int
            The number of seconds of data to store in one subdirectory. The
            timestamp of any subdirectory will be an integer multiple of this
            value.

        file_cadence_millisecs : int
            The number of milliseconds of data to store in each file. Note that
            an integer number of files must exactly span a subdirectory,
            implying::

                (subdir_cadence_secs*1000 % file_cadence_millisecs) == 0

        sample_rate_numerator : int
            Numerator of sample rate in Hz.

        sample_rate_denominator : int
            Denominator of sample rate in Hz.


        Other Parameters
        ----------------

        start : None | int | float | string, optional
            A value giving the time/index of the channel's first sample. When
            `ignore_tags` is False, 'rx_time' tags will be used to identify
            data gaps and skip the sample index forward appropriately (tags
            that refer to an earlier time will be ignored).
            If None or '' and `ignore_tags` is False, drop data until an
            'rx_time' tag arrives and sets the start time (a ValueError is
            raised if `ignore_tags` is True).
            If an integer, it is interpreted as a sample index given in the
            number of samples since the epoch (time_since_epoch*sample_rate).
            If a float, it is interpreted as a UTC timestamp (seconds since
            epoch).
            If a string, two forms are permitted:
                1) a string which can be evaluated to an integer/float and
                    interpreted as above,
                2) a time in ISO8601 format, e.g. '2016-01-01T16:24:00Z'

        ignore_tags : bool, optional
            If True, do not use 'rx_time' tags to set the sample index and do
            not write other tags as Digital Metadata.

        is_complex : bool, optional
            This parameter is only used when `dtype` is not complex.
            If True (the default), interpret supplied data as interleaved
            complex I/Q samples. If False, each sample has a single value.

        num_subchannels : int, optional
            Number of subchannels to write simultaneously. Default is 1.

        uuid_str : None | string, optional
            UUID string that will act as a unique identifier for the data and
            can be used to tie the data files to metadata. If None, a random
            UUID will be generated.

        center_frequencies : None | array_like of floats, optional
            List of subchannel center frequencies to include in initial
            metadata. If None, ``[0.0]*num_subchannels`` will be used.
            Subsequent center frequency metadata samples can be written using
            'rx_freq' stream tags.

        metadata : dict, optional
            Dictionary of additional metadata to include in initial Digital
            Metadata sample. Subsequent metadata samples can be written
            using 'metadata' stream tags, but all keys intended to be included
            should be set here first even if their values are empty.

        is_continuous : bool, optional
            If True, data will be written in continuous blocks. If False data
            will be written with gapped blocks. Fastest write/read speed is
            achieved with `is_continuous` True, `checksum` False, and
            `compression_level` 0 (all defaults).

        compression_level : int, optional
            0 for no compression (default), 1-9 for varying levels of gzip
            compression (1 == least compression, least CPU; 9 == most
            compression, most CPU).

        checksum : bool, optional
            If True, use HDF5 checksum capability. If False (default), no
            checksum.

        marching_periods : bool, optional
            If True, write a period to stdout for every subdirectory when
            writing.

        stop_on_skipped : bool, optional
            If True, stop writing when a sample would be skipped (such as from
            a dropped packet).

        debug : bool, optional
            If True, print debugging information.

        min_chunksize : None | int, optional
            Minimum number of samples to consume at once. This value can be
            used to adjust the sink's performance to reduce processing time.
            If None, a sensible default will be used.


        Notes
        -----

        By convention, this block sets the following Digital Metadata fields:

            uuid_str : string
                Value provided by the `uuid_str` argument.

            sample_rate_numerator : int
                Value provided by the `sample_rate_numerator` argument.

            sample_rate_denominator : int
                Value provided by the `sample_rate_denominator` argument.

            center_frequencies : list of floats with length `num_subchannels`
                Subchannel center frequencies as specified by
                `center_frequencies` argument and 'rx_freq' stream tags.

        Additional metadata fields can be set using the `metadata` argument and
        stream tags. Nested dictionaries are permitted and are helpful for
        grouping properties. For example, receiver-specific metadata is
        typically specified with a sub-dictionary using the 'receiver' field.


        This block acts on the following stream tags when `ignore_tags` is
        False:

            rx_time : (int secs, float frac) tuple
                Used to set the sample index from the given time since epoch.

            rx_freq : float
                Used to set the 'center_frequencies' value in the channel's
                Digital Metadata as described above.

            metadata : dict
                Used to populate additional (key, value) pairs in the channel's
                Digital Metadata. Any keys passed in 'metadata' tags should be
                included in the `metadata` argument at initialization to ensure
                that they always exist in the Digital Metadata.

        """
        dtype = np.dtype(dtype)
        # create structured dtype for interleaved samples if necessary
        if is_complex and (not np.issubdtype(dtype, np.complexfloating) and
                           not dtype.names):
            realdtype = dtype
            dtype = np.dtype([('r', realdtype), ('i', realdtype)])

        if num_subchannels == 1:
            in_sig = [dtype]
        else:
            in_sig = [(dtype, num_subchannels)]

        gr.sync_block.__init__(
            self,
            name="digital_rf_channel_sink",
            in_sig=in_sig,
            out_sig=None,
        )

        self._channel_dir = channel_dir
        self._channel_name = os.path.basename(channel_dir)
        self._dtype = dtype
        self._subdir_cadence_secs = subdir_cadence_secs
        self._file_cadence_millisecs = file_cadence_millisecs
        self._sample_rate_numerator = sample_rate_numerator
        self._sample_rate_denominator = sample_rate_denominator
        self._uuid_str = uuid_str
        self._ignore_tags = ignore_tags
        self._is_complex = is_complex
        self._num_subchannels = num_subchannels
        self._is_continuous = is_continuous
        self._compression_level = compression_level
        self._checksum = checksum
        self._marching_periods = marching_periods
        self._stop_on_skipped = stop_on_skipped
        self._debug = debug

        self._samples_per_second = (
            np.longdouble(np.uint64(sample_rate_numerator)) /
            np.longdouble(np.uint64(sample_rate_denominator))
        )

        if min_chunksize is None:
            self._min_chunksize = int(self._samples_per_second // 1000)
        else:
            self._min_chunksize = min_chunksize

        # reduce CPU usage by setting a minimum number of samples to handle
        # at once
        # (really want to set_min_noutput_items, but no way to do that from
        #  Python)
        self.set_output_multiple(self._min_chunksize)

        # will be None if start is None or ''
        self._start_sample = util.parse_identifier_to_sample(
            start, self._samples_per_second, None,
        )
        if self._start_sample is None:
            if self._ignore_tags:
                raise ValueError('Must specify start if ignore_tags is True.')
            # data without a time tag will be written starting at global index
            # of 0, i.e. the Unix epoch
            # we don't want to guess the start time because the user would
            # know better and it could obscure bugs by setting approximately
            # the correct time (samples in 1970 are immediately obvious)
            self._start_sample = 0
        self._next_rel_sample = 0

        # stream tags to read (in addition to rx_time, handled specially)
        if LooseVersion(gr.version()) >= LooseVersion('3.7.12'):
            self._stream_tag_translators = {
                pmt.intern('rx_freq'): translate_rx_freq,
                pmt.intern('metadata'): translate_metadata,
            }
        else:
            # USRP source in gnuradio < 3.7.12 has bad rx_freq tags, so avoid
            # trouble by ignoring rx_freq tags for those gnuradio versions
            self._stream_tag_translators = {
                pmt.intern('metadata'): translate_metadata,
            }

        # create metadata dictionary that will be updated and written whenever
        # new metadata is received in stream tags
        self._metadata = metadata.copy()
        if center_frequencies is None:
            center_frequencies = np.array([0.0]*self._num_subchannels)
        else:
            center_frequencies = np.ascontiguousarray(center_frequencies)
        self._metadata.update(
            # standard metadata by convention
            uuid_str='',
            sample_rate_numerator=self._sample_rate_numerator,
            sample_rate_denominator=self._sample_rate_denominator,
            center_frequencies=center_frequencies,
        )

        # create directories for RF data channel and metadata
        self._metadata_dir = os.path.join(self._channel_dir, 'metadata')
        if not os.path.exists(self._metadata_dir):
            os.makedirs(self._metadata_dir)

        # sets self._Writer, self._DMDWriter, and adds to self._metadata
        self._create_writer()

        # dict of metadata samples to be written, add for first sample
        # keys: absolute sample index for metadata
        # values: metadata dictionary to update self._metadata and then write
        self._md_queue = defaultdict(dict)
        self._md_queue[self._start_sample] = {}

    def _create_writer(self):
        # Digital RF writer
        self._Writer = DigitalRFWriter(
            self._channel_dir, self._dtype, self._subdir_cadence_secs,
            self._file_cadence_millisecs, self._start_sample,
            self._sample_rate_numerator, self._sample_rate_denominator,
            uuid_str=self._uuid_str, compression_level=self._compression_level,
            checksum=self._checksum, is_complex=self._is_complex,
            num_subchannels=self._num_subchannels,
            is_continuous=self._is_continuous,
            marching_periods=self._marching_periods,
        )
        # update UUID in metadata after parsing by DigitalRFWriter
        self._metadata.update(uuid_str=self._Writer.uuid)
        # Digital Metadata writer
        self._DMDWriter = DigitalMetadataWriter(
            metadata_dir=self._metadata_dir,
            subdir_cadence_secs=self._subdir_cadence_secs,
            file_cadence_secs=1,
            sample_rate_numerator=self._sample_rate_numerator,
            sample_rate_denominator=self._sample_rate_denominator,
            file_name='metadata',
        )

    def _read_tags(self, nsamples):
        """Read stream tags and set data blocks and metadata to write.

        Metadata from tags is added to the queue at ``self._md_queue``.

        """
        nread = self.nitems_read(0)
        md_queue = self._md_queue

        # continue writing at next continuous sample with start of block
        # unless overridden by a time tag
        data_blk_idxs = [0]
        data_rel_samples = [self._next_rel_sample]

        # read time tags
        time_tags = self.get_tags_in_window(
            0, 0, nsamples, pmt.intern('rx_time'),
        )
        # separate data into blocks to be written
        for tag in time_tags:
            offset = tag.offset
            tsec, tfrac, tidx = parse_time_pmt(
                tag.value, self._samples_per_second,
            )

            # index into data block for this tag
            bidx = offset - nread

            # get sample index relative to start
            sidx = tidx - self._start_sample

            # add new data block if valid and it indicates a gap
            prev_bidx = data_blk_idxs[-1]
            prev_sidx = data_rel_samples[-1]
            next_continuous_sample = prev_sidx + (bidx - prev_bidx)
            if sidx < next_continuous_sample:
                if self._debug:
                    errstr = (
                        '\n|{0}|rx_time tag @ sample {1}: {2}+{3} ({4})'
                        '\n INVALID: time cannot go backwards from index {5}.'
                        ' Skipping.'
                    ).format(
                        self._channel_name, offset, tsec, tfrac, tidx,
                        self._start_sample + next_continuous_sample,
                    )
                    sys.stdout.write(errstr)
                    sys.stdout.flush()
                continue
            elif sidx == next_continuous_sample:
                # don't create a new block because it's continuous
                continue
            else:
                # add new block to write based on time tag
                if self._debug:
                    tagstr = (
                        '\n|{0}|rx_time tag @ sample {1}: {2}+{3} ({4})'
                        '\n {5} dropped samples.'
                    ).format(
                        self._channel_name, offset, tsec, tfrac, tidx,
                        sidx - next_continuous_sample,
                    )
                    sys.stdout.write(tagstr)
                    sys.stdout.flush()
                if bidx == 0:
                    # override assumed continuous write
                    # data_blk_idxs[0] is already 0
                    data_rel_samples[0] = sidx
                else:
                    data_blk_idxs.append(bidx)
                    data_rel_samples.append(sidx)
                # reset metadata queue with only valid values
                for md_idx in md_queue.keys():
                    md_sidx = md_idx - self._start_sample
                    if next_continuous_sample <= md_sidx and md_sidx < sidx:
                        del md_queue[md_idx]
                # new metadata sample to help flag data skip
                md_queue.setdefault(sidx + self._start_sample, {})

        # read other tags by data block (so we know the sample index)
        for (bidx, bend), sidx in zip(
                pairwise(chain(data_blk_idxs, (nsamples,))),
                data_rel_samples,
        ):
            tags_by_offset = {}
            # read tags, translate to metadata dict, add to tag dict
            for tag_name, translator in self._stream_tag_translators.items():
                tags = self.get_tags_in_window(0, bidx, bend, tag_name)
                collect_tags_in_dict(tags, translator, tags_by_offset)
            # add tags to metadata sample dictionary
            for offset, tag_dict in tags_by_offset.items():
                mbidx = offset - nread
                # get the absolute sample index for the metadata sample
                msidx = (sidx + (mbidx - bidx)) + self._start_sample
                md_queue[msidx].update(tag_dict)

        return data_blk_idxs, data_rel_samples

    def work(self, input_items, output_items):
        in_data = input_items[0]
        nsamples = len(in_data)

        if not self._ignore_tags:
            # break data into blocks from time tags
            # get metadata from other tags and add to self._md_queue
            data_blk_idxs, data_rel_samples = self._read_tags(nsamples)
        else:
            # continue writing at next continuous sample with start of block
            data_blk_idxs = [0]
            data_rel_samples = [self._next_rel_sample]

        # make index lists into uint64 arrays
        data_rel_samples = np.array(data_rel_samples, dtype=np.uint64, ndmin=1)
        data_blk_idxs = np.array(data_blk_idxs, dtype=np.uint64, ndmin=1)

        # get any metadata samples to be written from queue
        if self._md_queue:
            md_samples, md_dict_updates = zip(*sorted(
                self._md_queue.items(), key=lambda x: x[0],
            ))
            md_samples = np.array(md_samples, dtype=np.uint64, ndmin=1)
            # fill out metadata to be written using stored metadata
            md_dicts = [
                # function updates self._metadata in-place, want copy for list
                # to preserve the updated values at that particular state
                recursive_dict_update(self._metadata, md_update).copy()
                for md_update in md_dict_updates
            ]
        else:
            md_samples = []
            md_dicts = []
        self._md_queue.clear()

        # check if skip occurs and break if so after writing continuous data
        if self._stop_on_skipped and len(data_blk_idxs) > 1:
            _py_rf_write_hdf5.rf_block_write(
                self._Writer._channelObj,
                in_data,
                data_rel_samples[:1],
                data_blk_idxs[:1],
            )
            last_rel_sample = (data_rel_samples[0] +
                               (data_blk_idxs[1] - data_blk_idxs[0]))
            last_sample = last_rel_sample + self._start_sample
            idx = np.searchsorted(md_samples, last_sample, 'right')
            for md_sample, md_dict in zip(md_samples[:idx], md_dicts[:idx]):
                self._DMDWriter.write(md_sample, md_dict)
            print("Stopping at skipped sample as requested.")
            # return WORK_DONE
            return -1

        # write metadata
        if md_dicts:
            self._DMDWriter.write(md_samples, md_dicts)

        # write data using block writer
        self._next_rel_sample = _py_rf_write_hdf5.rf_block_write(
            self._Writer._channelObj,
            in_data,
            data_rel_samples,
            data_blk_idxs,
        )

        return nsamples

    def stop(self):
        if self._Writer is not None:
            self._Writer.close()
        return super(digital_rf_channel_sink, self).stop()

    def get_debug(self):
        return self._debug

    def set_debug(self, debug):
        self._debug = debug

    def get_ignore_tags(self):
        return self._ignore_tags

    def set_ignore_tags(self, ignore_tags):
        self._ignore_tags = ignore_tags

    def get_stop_on_skipped(self):
        return self._stop_on_skipped

    def set_stop_on_skipped(self, stop_on_skipped):
        self._stop_on_skipped = stop_on_skipped


class digital_rf_sink(gr.hier_block2):
    """Sink block for writing Digital RF data."""

    def __init__(
        self, top_level_dir, channels, dtype, subdir_cadence_secs,
        file_cadence_millisecs, sample_rate_numerator, sample_rate_denominator,
        start=None, ignore_tags=False, is_complex=True, num_subchannels=1,
        uuid_str=None, center_frequencies=None, metadata={},
        is_continuous=True, compression_level=0,
        checksum=False, marching_periods=True, stop_on_skipped=True,
        debug=False, min_chunksize=None,
    ):
        """Write data in Digital RF format.

        This block is useful for writing multiple channels of data that have
        the same parameters. If different parameters for each channel are
        needed, use multiple `digital_rf_channel_sink` blocks.

        In addition to storing the input samples in Digital RF format, this
        block also populates each channel's accompanying Digital Metadata
        at the sample indices when the metadata changes or a data skip occurs.
        See the Notes section for details on what metadata is stored.

        Parameters
        ----------

        top_level_dir : string
            The top-level directory in which Digital RF channel directories
            will be created. It will be created if it does not exist.

        channels : list of strings | string
            List of channel names with length matching the number of channels
            to be written (and the number of inputs). Each channel name will
            be used as the directory name for the channel inside
            `top_level_dir`. If a string, a single channel will be written with
            that name.

        dtype : numpy.dtype | object to be cast by numpy.dtype()
            Object that gives the numpy dtype of the data to be written. This
            value is passed into ``numpy.dtype`` to get the actual dtype
            (e.g. ``numpy.dtype('>i4')``). Scalar types, complex types, and
            structured complex types with 'r' and 'i' fields of scalar types
            are valid.

        subdir_cadence_secs : int
            The number of seconds of data to store in one subdirectory. The
            timestamp of any subdirectory will be an integer multiple of this
            value.

        file_cadence_millisecs : int
            The number of milliseconds of data to store in each file. Note that
            an integer number of files must exactly span a subdirectory,
            implying::

                (subdir_cadence_secs*1000 % file_cadence_millisecs) == 0

        sample_rate_numerator : int
            Numerator of sample rate in Hz.

        sample_rate_denominator : int
            Denominator of sample rate in Hz.


        Other Parameters
        ----------------

        start : None | int | float | string, optional
            A value giving the time/index of the channel's first sample. When
            `ignore_tags` is False, 'rx_time' tags will be used to identify
            data gaps and skip the sample index forward appropriately (tags
            that refer to an earlier time will be ignored).
            If None or '' and `ignore_tags` is False, drop data until an
            'rx_time' tag arrives and sets the start time (a ValueError is
            raised if `ignore_tags` is True).
            If an integer, it is interpreted as a sample index given in the
            number of samples since the epoch (time_since_epoch*sample_rate).
            If a float, it is interpreted as a UTC timestamp (seconds since
            epoch).
            If a string, two forms are permitted:
                1) a string which can be evaluated to an integer/float and
                    interpreted as above,
                2) a time in ISO8601 format, e.g. '2016-01-01T16:24:00Z'

        ignore_tags : bool, optional
            If True, do not use 'rx_time' tags to set the sample index and do
            not write other tags as Digital Metadata.

        is_complex : bool, optional
            This parameter is only used when `dtype` is not complex.
            If True (the default), interpret supplied data as interleaved
            complex I/Q samples. If False, each sample has a single value.

        num_subchannels : int, optional
            Number of subchannels to write simultaneously. Default is 1.

        uuid_str : None | string, optional
            UUID string that will act as a unique identifier for the data and
            can be used to tie the data files to metadata. If None, a random
            UUID will be generated.

        center_frequencies : None | array_like of floats, optional
            List of subchannel center frequencies to include in initial
            metadata. If None, ``[0.0]*num_subchannels`` will be used.
            Subsequent center frequency metadata samples can be written using
            'rx_freq' stream tags.

        metadata : dict, optional
            Dictionary of additional metadata to include in initial Digital
            Metadata sample. Subsequent metadata samples can be written
            using 'metadata' stream tags, but all keys intended to be included
            should be set here first even if their values are empty.

        is_continuous : bool, optional
            If True, data will be written in continuous blocks. If False data
            will be written with gapped blocks. Fastest write/read speed is
            achieved with `is_continuous` True, `checksum` False, and
            `compression_level` 0 (all defaults).

        compression_level : int, optional
            0 for no compression (default), 1-9 for varying levels of gzip
            compression (1 == least compression, least CPU; 9 == most
            compression, most CPU).

        checksum : bool, optional
            If True, use HDF5 checksum capability. If False (default), no
            checksum.

        marching_periods : bool, optional
            If True, write a period to stdout for every subdirectory when
            writing.

        stop_on_skipped : bool, optional
            If True, stop writing when a sample would be skipped (such as from
            a dropped packet).

        debug : bool, optional
            If True, print debugging information.

        min_chunksize : None | int, optional
            Minimum number of samples to consume at once. This value can be
            used to adjust the sink's performance to reduce processing time.
            If None, a sensible default will be used.


        Notes
        -----

        By convention, this block sets the following Digital Metadata fields:

            uuid_str : string
                Value provided by the `uuid_str` argument.

            sample_rate_numerator : int
                Value provided by the `sample_rate_numerator` argument.

            sample_rate_denominator : int
                Value provided by the `sample_rate_denominator` argument.

            center_frequencies : list of floats with length `num_subchannels`
                Subchannel center frequencies as specified by
                `center_frequencies` argument and 'rx_rate' stream tags.

        Additional metadata fields can be set using the `metadata` argument and
        stream tags. Nested dictionaries are permitted and are helpful for
        grouping properties. For example, receiver-specific metadata is
        typically specified with a sub-dictionary using the 'receiver' field.


        This block acts on the following stream tags when `ignore_tags` is
        False:

            rx_time : (int secs, float frac) tuple
                Used to set the sample index from the given time since epoch.

            rx_freq : float
                Used to set the 'center_frequencies' value in the channel's
                Digital Metadata as described above.

            metadata : dict
                Used to populate additional (key, value) pairs in the channel's
                Digital Metadata. Any keys passed in 'metadata' tags should be
                included in the `metadata` argument at initialization to ensure
                that they always exist in the Digital Metadata.

        """
        options = locals()
        del options['self']
        del options['top_level_dir']
        del options['channels']

        self._top_level_dir = os.path.abspath(top_level_dir)
        if isinstance(channels, six.string_types):
            channels = [channels]
        self._channel_names = channels

        # make sinks for every channel
        self._channels = []
        for ch in self._channel_names:
            channel_dir = os.path.join(self._top_level_dir, ch)
            chsink = digital_rf_channel_sink(channel_dir, **options)
            self._channels.append(chsink)

        in_sig_dtypes = [sink.in_sig()[0] for sink in self._channels]
        in_sig = gr.io_signaturev(
            len(in_sig_dtypes), len(in_sig_dtypes),
            [s.itemsize for s in in_sig_dtypes],
        )
        out_sig = gr.io_signature(0, 0, 0)

        gr.hier_block2.__init__(
            self,
            name="digital_rf_sink",
            input_signature=in_sig,
            output_signature=out_sig,
        )

        for k, sink in enumerate(self._channels):
            self.connect((self, k), sink)

    def get_debug(self):
        return self._channels[0].get_debug()

    def set_debug(self, debug):
        for ch in self._channels:
            ch.set_debug(debug)

    def get_ignore_tags(self):
        return self._channels[0].ignore_tags()

    def set_ignore_tags(self, ignore_tags):
        for ch in self._channels:
            ch.set_ignore_tags(ignore_tags)

    def get_stop_on_skipped(self):
        return self._channels[0].get_stop_on_skipped()

    def set_stop_on_skipped(self, stop_on_skipped):
        for ch in self._channels():
            ch.set_stop_on_skipped(stop_on_skipped)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""test_digital_rf_hdf5.py is a script to test the digital_rf_hdf5 module

Duplicates a subset of the testing in test_rf_write_hdf5.c, then adds read testing

$Id$
"""
# standard python imports
import os
import time
import traceback
import glob
import shutil

# third party imports
import numpy
import h5py

# Millstone imports
import digital_rf

# base data
num_subchannels = 4
base_data = []
for i in range(100):
    if i < 50:
        base_data.append([2, 3] * num_subchannels)
    else:
        base_data.append([4, 6] * num_subchannels)

# create data in r/i to test using that to write
arr_data = numpy.ones((100, num_subchannels), dtype=[
                      ('r', numpy.int16), ('i', numpy.int16)])
arr_data[0:50, :]['r'] = 2
arr_data[50:, :]['r'] = 4
arr_data[0:50, :]['i'] = 3
arr_data[50:, :]['i'] = 6

# create single channel data in r/i to test using that to write
arr_data_single = numpy.ones(
    (100, 1), dtype=[('r', numpy.int16), ('i', numpy.int16)])
arr_data_single[0:50, :]['r'] = 2
arr_data_single[50:, :]['r'] = 4
arr_data_single[0:50, :]['i'] = 3
arr_data_single[50:, :]['i'] = 6


# simulate 10 samples of data followed by 10 sample gap
global_sample_arr = numpy.array(range(10), dtype=numpy.uint64) * 20
block_sample_arr = numpy.array(range(10), dtype=numpy.uint64) * 10

# constants
sample_rate_numerator = 200
sample_rate_denominator = 3
sample_rate = numpy.longdouble(sample_rate_numerator)/sample_rate_denominator
subdir_cadence_secs = 10
file_cadence_millisecs = 400
# start 2014-03-09 12:30:30 plus one sample
start_global_index = int(numpy.uint64(1394368230 * sample_rate)) + 1

# test get unix time
dt, picoseconds = digital_rf.get_unix_time(
    int(start_global_index), sample_rate_numerator, sample_rate_denominator)
print('For start_global_index=%i and sample_rate=%f, dt is %s and picoseconds is %i' % (start_global_index, sample_rate,
                                                                                        dt, picoseconds))

# set up top level directory
os.system("rm -rf /tmp/hdf5 ; mkdir /tmp/hdf5")
os.system("rm -rf /tmp/hdf52 ; mkdir /tmp/hdf52")

print("Test 0 - simple single write to multiple files, no compress, no checksum - channel 0")
os.system("rm -rf /tmp/hdf5/junk0 ; mkdir /tmp/hdf5/junk0")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk0", 'i4', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_0", 0, False, True, num_subchannels=num_subchannels)
data = numpy.array(base_data, numpy.int32)
result = data_object.rf_write(data)
data_object.close()
print("done test 0.1")

print("Test 0.1 - simple single write to multiple files using r/i struct layout, no compress, no checksum - channel 0.1")
os.system("rm -rf /tmp/hdf5/junk0.1 ; mkdir /tmp/hdf5/junk0.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk0.1", 'i2', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_0.1", 0, False, True, num_subchannels=num_subchannels)
result = data_object.rf_write(arr_data)
data_object.close()
print("done test 0.1")

print("Test 0.11 - simple single write with one subchannel to multiple files using r/i struct layout, no compress, no checksum - channel 0.1")
os.system("rm -rf /tmp/hdf5/junk0.11 ; mkdir /tmp/hdf5/junk0.11")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk0.11", 'i2', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_0.11", 0, False, True, num_subchannels=1)
result = data_object.rf_write(arr_data_single)
data_object.close()
print("done test 0.11")

print("Test 0.111 - gapped write with is_continuous==True with one subchannel to multiple files using r/i struct layout, no compress, no checksum - channel 0.1")
os.system("rm -rf /tmp/hdf5/junk0.111 ; mkdir /tmp/hdf5/junk0.111")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk0.111", 'i2', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_0.111", 0, False, True, num_subchannels=1)
result = data_object.rf_write_blocks(arr_data_single, [1, 5], [0, 2])
data_object.close()
print("done test 0.111")

print("Test 0.2 - read data from test 0.1 and write as channel 0.2")
os.system("rm -rf /tmp/hdf5/junk0.2 ; mkdir /tmp/hdf5/junk0.2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk0.2", 'i2', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_0.2", 0, False, True, num_subchannels=num_subchannels)
files = glob.glob('/tmp/hdf5/junk0.1/*/*.h5')
files.sort()
f = h5py.File(files[0])
read_data = f['rf_data'].value
result = data_object.rf_write(read_data)
data_object.close()
print("done test 0.2")

print("Test 1 - use complex 1 byte ints, no compress, no checksum - channel 1")
os.system("rm -rf /tmp/hdf5/junk1 ; mkdir /tmp/hdf5/junk1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1", 0, False, True, num_subchannels=num_subchannels)
data = numpy.array(base_data, numpy.int8)
data_object.rf_write(data)
data_object.close()
print(result)
print("done test 1")

print("Test 1.01 - use single 1 byte ints with no data gap, no compress, no checksum - channel 1.01")
os.system("rm -rf /tmp/hdf5/junk1.01 ; mkdir /tmp/hdf5/junk1.01")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1.01", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.01", 0, False, False, num_subchannels=1)
data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
data_object.rf_write(data)
data_object.close()
print("done test 1.01")

print("Test 1.1 - use single 1 byte ints with data gap, no compress, no checksum - channel 1.1")
os.system("rm -rf /tmp/hdf5/junk1.1 ; mkdir /tmp/hdf5/junk1.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1.1", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.1", 0, False, False, num_subchannels=1)
data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
data_object.rf_write(data)
data_object.rf_write(data, 120)
data_object.close()
print("done test 1.1")

print("Test 1.2 - use single 1 byte ints with no gaps across multiple subdirectories and top_level dirs, no compress, no checksum - channel 1.1")
os.system("rm -rf /tmp/hdf5/junk1.2 ; mkdir /tmp/hdf5/junk1.2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1.2", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.2", 0, False, False, num_subchannels=1)
data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
for i in range(100):
    data_object.rf_write(data)
data_object.close()
print('now use top level dir /tmp/hdf52')
os.system("rm -rf /tmp/hdf52/junk1.2 ; mkdir /tmp/hdf52/junk1.2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf52/junk1.2", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index + 100 * 100,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.2", 0, False, False, num_subchannels=1)
"""# buggy version
data_object = digital_rf.DigitalRFWriter("/tmp/hdf52/junk1.2", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index + 50*100,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.2", 0, False, False);   """
data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
for i in range(100):
    data_object.rf_write(data)
data_object.close()
print("done test 1.2")

print("Test 2 - use 2 byte ints with data gap, level 1 compress, but no checksum - channel 2")
os.system("rm -rf /tmp/hdf5/junk2 ; mkdir /tmp/hdf5/junk2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk2", 'i2', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_2", 1, False, True, num_subchannels=num_subchannels)
data = numpy.array(base_data, numpy.int16)
data_object.rf_write(data)
data_object.rf_write(data, 120)
data_object.close()
print("done test 2")

print("Test 3 - use complex float, no compress, no checksum - channel 3")
os.system("rm -rf /tmp/hdf5/junk3 ; mkdir /tmp/hdf5/junk3")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk3", 'f', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_3", 0, False, True, num_subchannels=num_subchannels)
data = numpy.array(base_data, numpy.float32)
data_object.rf_write(data)
data_object.close()
print(result)
print("done test 3")

print("Test 3.1 - use complex double, no compress, no checksum - channel 3")
os.system("rm -rf /tmp/hdf5/junk3.1 ; mkdir /tmp/hdf5/junk3.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk3.1", 'd', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_3.1", 0, False, True, num_subchannels=num_subchannels)
data = numpy.array(base_data, numpy.double)
data_object.rf_write(data)
data_object.close()
print(result)
print("done test 3.1")

print("Test 3.2 - use complex double in complex form, no compress, no checksum - channel 3")
os.system("rm -rf /tmp/hdf5/junk3.2 ; mkdir /tmp/hdf5/junk3.2")
dbl_data = numpy.array(base_data, numpy.double)
data = numpy.zeros((dbl_data.shape[0], dbl_data.shape[1]/2), numpy.complex128)
data.real = dbl_data[:, ::2]
data.imag = dbl_data[:, 1::2]
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk3.2", data.dtype, subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_3.2", 0, False, True, num_subchannels=num_subchannels)
data_object.rf_write(data)
data_object.close()
print(result)
print("done test 3.2")

print("Test 4.1 - use single 8 byte ints with 10 on/10 missing blocks, both compress (level 6) and checksum - channel 4.1")
os.system("rm -rf /tmp/hdf5/junk4.1 ; mkdir /tmp/hdf5/junk4.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk4.1", 'i8', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_2", 1, False, True, num_subchannels=num_subchannels,
                                                 is_continuous=False)
data = numpy.array(base_data, numpy.int64)
for i in range(100):
    data_object.rf_write_blocks(data, global_sample_arr, block_sample_arr)
    global_sample_arr += 205
data_object.close()

print('Now write more of this same test 4.1 to top level directory /tmp/hdf52')
os.system("rm -rf /tmp/hdf52/junk4.1 ; mkdir /tmp/hdf52/junk4.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf52/junk4.1", 'i8', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                                 sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_2", 1, False, True, num_subchannels=num_subchannels,
                                                 is_continuous=False)
data = numpy.array(base_data, numpy.int64)
for i in range(100):
    data_object.rf_write_blocks(data, global_sample_arr, block_sample_arr)
    global_sample_arr += 205
data_object.close()
print("done test 4.1")

# sleep for 4 seconds to make sure system knows all files closed
time.sleep(4)

# read
t = time.time()
# test relative paths
pwd = os.getcwd()
os.chdir('/tmp')
testReadObj = digital_rf.DigitalRFReader(['hdf5', 'hdf52'])
print('init took %f' % (time.time() - t))

channels = testReadObj.get_channels()
print(channels)

print('working on channel4.1')
start_index, end_index = testReadObj.get_bounds('junk4.1')
print((start_index, end_index))
print('calling get_continuous_blocks')
block_start = start_global_index + 205 * 100 - 67
block_end = block_start + 204
cont_data_dict = testReadObj.get_continuous_blocks(
    block_start, block_end, 'junk4.1')
print(cont_data_dict)
start_sample = cont_data_dict.keys()[0]
sample_len = cont_data_dict[start_sample]
print((start_sample, sample_len))

print('checking get_properties - global values')
d = testReadObj.get_properties('junk4.1')
keys = d.keys()
keys.sort()
for key in keys:
    print('%s: %s' % (str(key), str(d[key])))

print('checking get_properties at set sammple %i' % (start_sample))
d = testReadObj.get_properties('junk4.1', start_sample)
keys = d.keys()
keys.sort()
for key in keys:
    print('%s: %s' % (str(key), str(d[key])))

# normal read
result = testReadObj.read_vector_raw(start_sample, sample_len, 'junk4.1')
print(len(result))
print(result)
print(result.dtype)
try:
    # read too far
    result = testReadObj.read_vector_raw(
        start_sample, sample_len + 1, 'junk4.1')
    raise ValueError, 'whoops - no error when one expected!!!!!'
except IOError:
    traceback.print_exc()
    print('got expected error')
# read one less than block - should be okay
t = time.time()
result = testReadObj.read_vector_raw(start_sample, sample_len - 1, 'junk4.1')
print('read took %f' % (time.time() - t))
print(len(result))
print(result)

print('working on channel1.2 - large continuous data')
start_index, end_index = testReadObj.get_bounds('junk1.2')
print((start_index, end_index))
print(testReadObj.get_continuous_blocks(start_index, end_index, 'junk1.2'))
t = time.time()
result = testReadObj.read_vector_raw(
    start_index, end_index - start_index, 'junk1.2')
print('read took %f' % (time.time() - t))
print(len(result))

print('print all channel rf file metadata for junk1.2')
this_dict = testReadObj.get_properties('junk1.2')
keys = this_dict.keys()
keys.sort()
for key in keys:
    print((key, this_dict[key]))

metadata_files = glob.glob(os.path.join(pwd, 'metadata*.000.h5'))
for metadata_file in metadata_files:
    shutil.copy(metadata_file, '/tmp/hdf5/junk1.2')

print('Test of read_vector from single subchannel channel')
start_index, end_index = testReadObj.get_bounds('junk0.11')
result = testReadObj.read_vector(
    start_index, end_index - start_index, 'junk0.11')
print('result')
print(result)
print('result.shape is %s' % (str(result.shape)))
print(result.dtype)

print('Test of read_vector from 4 subchannel channel')
start_index, end_index = testReadObj.get_bounds('junk0')
result = testReadObj.read_vector(start_index, end_index - start_index, 'junk0')
print('result')
print(result)
print('result.shape is %s' % (str(result.shape)))
print(result.dtype)

print('Test of read_vector from 4 subchannel float channel')
start_index, end_index = testReadObj.get_bounds('junk3')
result = testReadObj.read_vector(start_index, end_index - start_index, 'junk3')
print('result')
print(result)
print('result.shape is %s' % (str(result.shape)))
print(result.dtype)

print('Test of read_vector from 4 subchannel double channel')
start_index, end_index = testReadObj.get_bounds('junk3.1')
result = testReadObj.read_vector(
    start_index, end_index - start_index, 'junk3.1')
print('result')
print(result)
print('result.shape is %s' % (str(result.shape)))
print(result.dtype)

print('Test of read_vector from 4 subchannel double channel v2')
start_index, end_index = testReadObj.get_bounds('junk3.2')
result = testReadObj.read_vector(
    start_index, end_index - start_index, 'junk3.2')
print('result')
print(result)
print('result.shape is %s' % (str(result.shape)))
print(result.dtype)

print('Verify read_vector sets all imaginary values to zero with single valued channel')
start_index, end_index = testReadObj.get_bounds('junk1.01')
result = testReadObj.read_vector(
    start_index, end_index - start_index, 'junk1.01')
if len(numpy.nonzero(result.imag.flatten())[0]) > 0:
    raise ValueError, 'Got imaginary part when not expected'

print('Overall test passed')



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""benchmark_rf_read_hdf5.py is a script to benchmark reading the Hdf5 files produced by benchmark_rf_write_hdf5.py

$Id$
"""
# standard python imports
import os
import sys
import time

# Millstone imports
import digital_rf

# constants
WRITE_BLOCK_SIZE = 1000
FILE_SAMPLES = 1000000
N_WRITES = int(1e9 / WRITE_BLOCK_SIZE)


def test_read(channel_name, test_read_obj):
    """test_read measures the speed of reading all the data back into numpy arrays for
    a given channel_name and digital_rf.DigitalRFReader object
    """
    if sys.platform == 'darwin':
        os.system('purge')
    t2 = time.time()
    start_index, end_index = test_read_obj.get_bounds(channel_name)
    print('get_bounds returned %i - %i and took %f seconds' %
          (start_index, end_index, time.time() - t2))
    t2 = time.time()
    result = test_read_obj.get_continuous_blocks(
        start_index, end_index, channel_name)
    print('get_continuous_blocks returned <%s> and took %f seconds' %
          (str(result), time.time() - t2))

    next_sample = start_index
    t2 = time.time()
    count = 0
    while next_sample < end_index - FILE_SAMPLES:
        arr = test_read_obj.read(
            next_sample, next_sample + (FILE_SAMPLES - 1), channel_name)
        key = arr.keys()[0]
        if len(arr[key]) != FILE_SAMPLES:
            raise IOError, '%i != %i' % (len(arr), FILE_SAMPLES)
        next_sample += FILE_SAMPLES
        count += 1
        if count % 100 == 0:
            print('%i out of 1000' % (count))
    seconds = time.time() - t2
    speedMB = (count * FILE_SAMPLES * 4) / (1.0E6 * seconds)
    print('Total read time %i seconds, speed %1.2f MB/s' %
          (int(seconds), speedMB))

t = time.time()
test_read_obj = digital_rf.DigitalRFReader('/tmp/benchmark')
print('metadata analysis took %f seconds' % (time.time() - t))

print("\nTest 0 - read Hdf5 files with no compress, no checksum - channel name = junk0")
test_read('junk0', test_read_obj)


print("\nTest 1 - read Hdf5 files with no compress, no checksum, chunked - channel name = junk1")
test_read('junk1', test_read_obj)


print("\nTest 2 -read Hdf5 files with no compress, but with level 9 checksum - channel name = junk2")
test_read('junk2', test_read_obj)


print("\nTest 3 - read Hdf5 files with compress, and with level 9 checksum - channel name = junk3")
test_read('junk3', test_read_obj)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""example of digital_rf_metadata.write_digital_metadata

Now writes data into two levels of dictionaries/groups.  API allow any finite number
of levels.

$Id$
"""
# standard python imports
import os

# third party imports
import numpy

# Millstone imports
import digital_rf

metadata_dir = '/tmp/test_metadata'
subdirectory_cadence_seconds = 3600
file_cadence_seconds = 60
samples_per_second_numerator = 10
samples_per_second_denominator = 9
file_name = 'rideout'
stime = 1447082580

os.system('mkdir %s' % (metadata_dir))
os.system('rm -r %s/*' % (metadata_dir))

obj = digital_rf.DigitalMetadataWriter(metadata_dir, subdirectory_cadence_seconds, file_cadence_seconds,
                                       samples_per_second_numerator, samples_per_second_denominator,
                                       file_name)
print('first create okay')

data_dict = {}
start_idx = int(numpy.uint64(stime * obj.get_samples_per_second()))
# To save an array of data, make sure the first axis has the same length
# as the samples index
idx_arr = numpy.arange(70, dtype=numpy.int64) + start_idx

int_data = numpy.arange(70, dtype=numpy.int32)
data_dict['int_data'] = int_data
# can even do multi-dimensional arrays!
int_mat = numpy.arange(70*3*2, dtype=numpy.int32).reshape(70, 3, 2)
data_dict['int_mat'] = int_mat
# These single dimensional arrays will save each individual elements to each
# time.
float_data = numpy.arange(70, dtype=numpy.float32)
data_dict['float_data'] = float_data
complex_data = numpy.arange(70, dtype=numpy.complex64)
data_dict['complex_data'] = complex_data
single_int = 5
data_dict['single_int'] = numpy.int32(single_int)
single_float = 6.0
data_dict['single_float'] = numpy.float64(single_float)
single_complex = 7.0 + 8.0j
data_dict['single_complex'] = numpy.complex(single_complex)

# now create subdirectories
sub_dict = {}
sub_dict['single_int'] = numpy.int32(single_int)
sub_dict['single_float'] = numpy.float64(single_float)
sub_dict['single_complex'] = numpy.complex(single_complex)
level2_dict = {}  # embed yey another level
level2_dict['single_float'] = numpy.float64(single_float)
sub_dict['level2'] = level2_dict


data_dict['sub_system'] = sub_dict

# complex python object
n = numpy.ones((10, 4), dtype=numpy.float64)
n[5, :] = 17.0
data_dict['numpy_obj'] = [n for i in range(70)]


obj.write(idx_arr, data_dict)
print('first write_metadata okay')

# write same data again after incrementating inx
idx_arr += 70

del(obj)
obj = digital_rf.DigitalMetadataWriter(metadata_dir, subdirectory_cadence_seconds, file_cadence_seconds,
                                       samples_per_second_numerator, samples_per_second_denominator,
                                       file_name)
print('second create okay')


obj.write(idx_arr, data_dict)
print('second write_metadata okay')



"""Digital RF Python package."""
try:
    from ._version import __version__
except ImportError:
    __version__ = None
from .digital_metadata import *
from .digital_rf_hdf5 import *
from . import list_drf
from .list_drf import ilsdrf, lsdrf
from . import util
try:
    from . import mirror
    from . import ringbuffer
    from . import watchdog_drf
except ImportError:
    # if no watchdog package, these fail to import, so just ignore
    pass



"""Digital RF module for GNU Radio."""

try:
    from ._version import __version__
except ImportError:
    __version__ = None
from .digital_rf_source import digital_rf_source, digital_rf_channel_source
from .digital_rf_sink import digital_rf_sink, digital_rf_channel_sink



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Create pseudorandom-coded waveform files for sounding.

See the following paper for a description and application of meteor radar using
pseudorandom codes:

Vierinen, J., Chau, J. L., Pfeffer, N., Clahsen, M., and Stober, G.,
Coded continuous wave meteor radar, Atmos. Meas. Tech., 9, 829-839,
doi:10.5194/amt-9-829-2016, 2016.

"""
import math
from argparse import ArgumentParser

import numpy
import scipy.signal


# seed is a way of reproducing the random code without
# having to store all actual codes. the seed can then
# act as a sort of station_id.
def create_pseudo_random_code(clen=10000, seed=0):
    numpy.random.seed(seed)
    phases = numpy.array(
        numpy.exp(1.0j * 2.0 * math.pi * numpy.random.random(clen)),
        dtype=numpy.complex64,
    )
    return(phases)


# oversample a phase code by a factor of rep
def rep_seq(x, rep=10):
    L = len(x) * rep
    res = numpy.zeros(L, dtype=x.dtype)
    idx = numpy.arange(len(x)) * rep
    for i in numpy.arange(rep):
        res[idx + i] = x
    return(res)


#
# lets use 0.1 s code cycle and coherence assumption
# our transmit bandwidth is 100 kHz, and with a 10e3 baud code,
# that is 0.1 seconds per cycle as a coherence assumption.
# furthermore, we use a 1 MHz bandwidth, so we oversample by a factor of 10.
#
def waveform_to_file(
    station=0, clen=10000, oversample=10, filter_output=False,
):
    a = rep_seq(
        create_pseudo_random_code(clen=clen, seed=station),
        rep=oversample,
    )
    if filter_output:
        w = numpy.zeros([oversample * clen], dtype=numpy.complex64)
        fl = (int(oversample + (0.1 * oversample)))
        w[0:fl] = scipy.signal.blackmanharris(fl)
        aa = numpy.fft.ifft(numpy.fft.fft(w) * numpy.fft.fft(a))
        a = aa / numpy.max(numpy.abs(aa))
        a = numpy.array(a, dtype=numpy.complex64)
        a.tofile('code-l%d-b%d-%06df.bin' % (clen, oversample, station))
    else:
        a.tofile('code-l%d-b%d-%06d.bin' % (clen, oversample, station))


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument(
        '-l', '--length', type=int, default=10000,
        help='''Code length (number of bauds). (default: %(default)s)''',
    )
    parser.add_argument(
        '-b', '--oversampling', type=int, default=10,
        help='''Oversampling factor (number of samples per baud).
                (default: %(default)s)''',
    )
    parser.add_argument(
        '-s', '--station', type=int, default=0,
        help='''Station ID (seed). (default: %(default)s)''',
    )
    parser.add_argument(
        '-f', '--filter', action='store_true',
        help='''Filter waveform with Blackman-Harris window.
                (default: %(default)s)''',
    )

    op = parser.parse_args()

    waveform_to_file(
        station=op.station, clen=op.length, oversample=op.oversampling,
        filter_output=op.filter,
    )



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""example1.py used used to test bug relating to need for setting c array order

$Id$
"""
# standard python imports
import datetime
import time

# third party imports
import numpy

# Millstone imports
import digital_rf


data = numpy.array([[0.00074581,  0.00997215],
                    [-0.00222403, -0.00974955],
                    [0.00365252,  0.00930909],
                    [-0.00499937, -0.00866062],
                    [0.00623448,  0.00781865],
                    [-0.00733026, -0.006802],
                    [0.00826225,  0.0056334],
                    [-0.00900963, -0.00433896],
                    [0.00955571,  0.00294761]])

print(data)
print(data.shape)

voltage = numpy.array([0.00074581 + 0.00997215j,
                       -0.00222403 + -0.00974955j,
                       0.00365252 + 0.00930909j,
                       -0.00499937 + -0.00866062j,
                       0.00623448 + 0.00781865j,
                       -0.00733026 + -0.006802j,
                       0.00826225 + 0.0056334j,
                       -0.00900963 + -0.00433896j,
                       0.00955571 + 0.00294761j], dtype=numpy.complex)


data_ri = []
data_ri.append(numpy.real(voltage))
data_ri.append(numpy.imag(voltage))
data_ri = numpy.transpose(numpy.array(data_ri))

print(data_ri)
print(data_ri.shape)

# constants
save_dir_data = '/tmp/test/chb'
sample_rate_numerator = int(1.0E6)
sample_rate_denominator = 1
sample_rate = numpy.longdouble(sample_rate_numerator) / sample_rate_denominator
start_time = datetime.datetime(2017, 1, 1)
t = int(numpy.uint64(time.mktime(start_time.timetuple()) * sample_rate))


drf_out2 = digital_rf.DigitalRFWriter(save_dir_data, 'f', 3600, 1000, t, sample_rate_numerator, sample_rate_denominator,
                                      'drf_dummy_data_v0.1.py', 0, False, True, 1,
                                      False, False)
drf_out2.rf_write(data_ri.astype('f4'))
drf_out2.close()

testReadObj = digital_rf.DigitalRFReader(['/tmp/test'])
start_index, end_index = testReadObj.get_bounds('chb')
blocks = testReadObj.get_continuous_blocks(start_index, end_index, 'chb')
for measurement_index in blocks:
    data = testReadObj.read_vector(
        int(measurement_index), int(blocks[measurement_index]), 'chb')
    print(data)
    print(data.dtype)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""benchmark_rf_write_hdf5.py is a script to to run the same test as in benchmark_rf_write_hdf5.c,
except in C.  All tests single subchannel.

$Id$
"""
# standard python imports
import os
import time

# third party imports
import numpy

# Millstone imports
import digital_rf

# constants
WRITE_BLOCK_SIZE = 1000
N_WRITES = int(1e9 / WRITE_BLOCK_SIZE)
SAMPLE_RATE_NUMERATOR = int(1E9)
SAMPLE_RATE_DENOMINATOR = 1
sample_rate = (numpy.longdouble(numpy.uint64(SAMPLE_RATE_NUMERATOR)) /
               numpy.longdouble(numpy.uint64(SAMPLE_RATE_DENOMINATOR)))
subdir_cadence_secs = 3600
file_cadence_millisecs = 10

# start 2014-03-09 12:30:30 plus one sample
start_global_index = int(numpy.uint64(1394368230 * sample_rate)) + 1


# data to write
data_int16 = numpy.zeros((WRITE_BLOCK_SIZE, 2), dtype='i2')
# make random
for i in range(WRITE_BLOCK_SIZE):
    j = i * 2
    k = i * 2 + 1
    data_int16[i][0] = (j % 32768) * (j + 8192) * (j % 13)
    data_int16[i][1] = (k % 32768) * (k + 8192) * (k % 13)

print('creating top level dir /tmp/benchmark')
os.system("rm -rf /tmp/benchmark ; mkdir /tmp/benchmark")

print("Test 0 - simple single write to multiple files, no compress, no checksum - channel 0")
os.system("rm -rf /tmp/benchmark/junk0 ; mkdir /tmp/benchmark/junk0")
print("Start writing")
channelObj = digital_rf.DigitalRFWriter('/tmp/benchmark/junk0', 'i2', subdir_cadence_secs, file_cadence_millisecs,
                                        start_global_index, SAMPLE_RATE_NUMERATOR, SAMPLE_RATE_DENOMINATOR, 'Fake_uuid', 0, False)
t = time.time()
for i in range(N_WRITES):
    channelObj.rf_write(data_int16)
channelObj.close()
seconds = time.time() - t
speedMB = (N_WRITES * 4 * WRITE_BLOCK_SIZE) / (1.0E6 * seconds)
print('Total time %i seconds, speed %1.2f MB/s' % (int(seconds), speedMB))

print("Test 1 - simple single write to multiple files, no compress, no checksum, chunked - channel 1")
os.system("rm -rf /tmp/benchmark/junk1 ; mkdir /tmp/benchmark/junk1")
print("Start writing")
channelObj = digital_rf.DigitalRFWriter('/tmp/benchmark/junk1', 'i2', subdir_cadence_secs, file_cadence_millisecs,
                                        start_global_index, SAMPLE_RATE_NUMERATOR, SAMPLE_RATE_DENOMINATOR, 'Fake_uuid', 0, False, is_continuous=False)
t = time.time()
for i in range(N_WRITES):
    channelObj.rf_write(data_int16)
channelObj.close()
seconds = time.time() - t
speedMB = (N_WRITES * 4 * WRITE_BLOCK_SIZE) / (1.0E6 * seconds)
print('Total time %i seconds, speed %1.2f MB/s' % (int(seconds), speedMB))


print("Test 2 - simple single write to multiple files, no compress, with checksum - channel 2")
os.system("rm -rf /tmp/benchmark/junk2 ; mkdir /tmp/benchmark/junk2")
print("Start writing")
channelObj = digital_rf.DigitalRFWriter('/tmp/benchmark/junk2', 'i2', subdir_cadence_secs, file_cadence_millisecs,
                                        start_global_index, SAMPLE_RATE_NUMERATOR, SAMPLE_RATE_DENOMINATOR, 'Fake_uuid', 0, True)
t = time.time()
for i in range(N_WRITES):
    channelObj.rf_write(data_int16)
channelObj.close()
seconds = time.time() - t
speedMB = (N_WRITES * 4 * WRITE_BLOCK_SIZE) / (1.0E6 * seconds)
print('Total time %i seconds, speed %1.2f MB/s' % (int(seconds), speedMB))


print("Test 3 - simple single write to multiple files, compress to level 9, with checksum - channel 3")
os.system("rm -rf /tmp/benchmark/junk3 ; mkdir /tmp/benchmark/junk3")
print("Start writing")
channelObj = digital_rf.DigitalRFWriter('/tmp/benchmark/junk3', 'i2', subdir_cadence_secs, file_cadence_millisecs,
                                        start_global_index, SAMPLE_RATE_NUMERATOR, SAMPLE_RATE_DENOMINATOR, 'Fake_uuid', 9, True)
t = time.time()
for i in range(N_WRITES):
    channelObj.rf_write(data_int16)
channelObj.close()
seconds = time.time() - t
speedMB = (N_WRITES * 4 * WRITE_BLOCK_SIZE) / (1.0E6 * seconds)
print('Total time %i seconds, speed %1.2f MB/s' % (int(seconds), speedMB))



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Record beacon signals of specified satellites in Digital RF.

Satellite and recording parameters are specified in .ini configuration files.
Example configurations are included along with this script.

"""

import sys
import os
import traceback
import subprocess
import string
import time
import datetime
import dateutil.parser
import pytz
import math
import numpy
import ephem
import optparse
import ConfigParser

from digital_rf import DigitalMetadataWriter

class ExceptionString(Exception):
    """ Simple exception handling string """
    def __str__(self):
          return repr(self.args[0])


def doppler_shift(frequency, relativeVelocity):
    """
    DESCRIPTION:
        This function calculates the doppler shift of a given frequency when actual
        frequency and the relative velocity is passed.
        The function for the doppler shift is f' = f - f*(v/c).
    INPUTS:
        frequency (float)        = satlitte's beacon frequency in Hz
        relativeVelocity (float) = Velocity at which the satellite is moving
                                   towards or away from observer in m/s
    RETURNS:
        Param1 (float)           = The frequency experienced due to doppler shift in Hz
    AFFECTS:
        None
    EXCEPTIONS:
        None
    DEPENDENCIES:
        ephem.Observer(...), ephem.readtle(...)
    Note: relativeVelocity is positive when moving away from the observer
          and negative when moving towards
    """
    return  (frequency - frequency * (relativeVelocity/3e8))


def satellite_rise_and_set(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, startDate):
    """
    DESCRIPTION:
        This function take in the observers latitude, longitude, elevation as well
        as the object's name and TLE line 1 and 2 to calculate the next closest rise
        and set times from the given start date. Returns an array of values.
    INPUTS:
        obsLat (string)                  = Latitude of Observer in degrees represented as strings
        obsLong (string)                 = Longitude of Observer in degrees represented as strings
        obsElev (float)                  = Elevation of Observer in meters
        objName (string)                 = Name of the satellite
        tle1 (string)                    = First line of TLE
        tle2 (string)                    = Second line of TLE
        startDate (string or ephem.date) = The date from which next closest rise and set are to be
                                           found in radians that print as degrees

    RETURNS:
        Param1 (ephem.date)              = Rise time of satellite in 'yyyy/mm/dd hh:mm:ss'
        Param2 (ephem.date)              = Half time between rise and set in 'yyyy/mm/dd hh:mm:ss'
        Param3 (ephem.date)              = Set time of satellite in 'yyyy/mm/dd hh:mm:ss'
    AFFECTS:
        None
    EXCEPTIONS:
        None
    DEPENDENCIES:
        ephem.Observer(...), ephem.readtle(...)
    """
    obsLoc = ephem.Observer()
    obsLoc.lat = obsLat
    obsLoc.long = obsLong
    obsLoc.elev = obsElev
    obsLoc.date = startDate

    if opt.debug:
        print 'dbg location: ', obsLoc

        print 'dbg tle1: ', tle1
        print 'dbg tle2: ', tle2

    satObj = ephem.readtle(objName,tle1,tle2)

    if opt.debug:
        print 'dbg object: ', satObj

    satObj.compute(obsLoc) # computes closest next rise time to given date

    pinfo = obsLoc.next_pass(satObj)

    return (pinfo[0], pinfo[2], pinfo[4])



def satellite_values_at_time(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, date):
    """
    DESCRIPTION:
        This function take in the observers latitude, longitude, elevation as well
        as the object's name and TLE line 1 and 2 to calculate various values
        at a given date. It returns an array of values
    INPUTS:
        obsLat (string)                  = Latitude of Observer in degrees represented as strings
        obsLong (string)                 = Longitude of Observer in degrees represented as strings
        obsElev (float)                  = Elevation of Observer in meters
        objName (string)                 = Name of the satellite
        tle1 (string)                    = First line of TLE
        tle2 (string)                    = Second line of TLE
        date (string or ephem.date)      = The date from which next closest rise and set are to be
                                           found in radians that print as degrees

    RETURNS:
        Param1 (ephem.angle)             = satellite's latitude in radians that print as degrees
        Param2 (ephem.angle)             = satellite's longitude in radians that print as degrees
        Param3 (float)                   = satellite's current range from the observer in meters
        Param4 (float)                   = satellite's current range_velocity from the observer in m/s
        Param5 (ephem.angle)             = satellite's current azimuth in radians that print as degrees
        Param6 (ephem.angle)             = satellite's current altitude in radians that print as degrees
        Param7 (ephem.angle)             = satellite's right ascention in radians that print as hours of arc
        Param8 (ephem.angle)             = satellite's declination in radians that print as degrees
        Param9 (float)                   = satellite's elevation in meters from sea level
    AFFECTS:
        None
    EXCEPTIONS:
        None
    DEPENDENCIES:
        ephem.Observer(...), ephem.readtle(...)
    """
    obsLoc = ephem.Observer()
    obsLoc.lat = obsLat
    obsLoc.long = obsLong
    obsLoc.elev = obsElev
    obsLoc.date = date

    satObj = ephem.readtle(objName,tle1,tle2)
    satObj.compute(obsLoc)

    if opt.debug:
        print "\tLatitude: %s, Longitude %s, Range: %gm, Range Velocity: %gm/s" % (satObj.sublat, satObj.sublong, satObj.range, satObj.range_velocity)
        print "\tAzimuth: %s, Altitude: %s, Elevation: %gm" % ( satObj.az, satObj.alt, satObj.elevation)
        print "\tRight Ascention: %s, Declination: %s" % (satObj.ra, satObj.dec)

    return (satObj.sublat, satObj.sublong, satObj.range, satObj.range_velocity, satObj.az, satObj.alt, satObj.ra, satObj.dec, satObj.elevation)

def max_satellite_bandwidth(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, startDate, endDate, interval, beaconFreq):
    """
    DESCRIPTION:
        The function calls the satellite_bandwidth function over and over for however many rises and sets occur
        during the [startDate, endDate]. The max bandwidth is then returned.
    INPUTS:
        obsLat (string)                  = Latitude of Observer in degrees represented as strings
        obsLong (string)                 = Longitude of Observer in degrees represented as strings
        obsElev (float)                  = Elevation of Observer in meters
        objName (string)                 = Name of the satellite
        tle1 (string)                    = First line of TLE
        tle2 (string)                    = Second line of TLE
        startDate (string or ephem.date) = The date/time at which to find the first cycle
        endDate (string or ephem.date)   = The date/time at which to stop looking for a cycle
        interval (float)                = The rate at which to sample during one rise/set cycle same format
                                           as time
        beaconFreq (float)               = The frequency of the beacon

    RETURNS:
        Param1 (float)                   = The max bandwidth of the satellite in the given range
                                           of start and end dates
    AFFECTS:
        None
    EXCEPTIONS:
        None
    DEPENDENCIES:
        None
    """
    maxBandwidth = 0
    (satRise,satTransit,satSet) = satellite_rise_and_set(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, startDate)
    if(satRise == satTransit  == satSet):
        return 0

    while(satRise < endDate):
        (objBandwidth, shiftedFrequencies) = satellite_bandwidth(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, satRise, satSet, interval, beaconFreq)
        if(objBandwidth > maxBandwidth):
            maxBandwidth = objBandwidth
        (satRise,satTransit, satSet) = satellite_rise_and_set(opt,obsLat, obsLong, obsElev, objName, tle1, tle2, satSet+ephem.minute*5.0)
        #print "Name: %s, Rise Time: %s, Transit Time: %s, Set Time: %s" % (objName, ephem.date(satRise-ephem.hour*4.0), ephem.date(satTransit-ephem.hour*4.0), ephem.date(satSet-ephem.hour*4.0))

    return maxBandwidth

def satellite_bandwidth(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, satRise, satSet, interval, beaconFreq):
    """
    DESCRIPTION:
        The function finds the bandwidth of a satellite pass
    INPUTS:
        obsLat (string)                  = Latitude of Observer in degrees represented as strings
        obsLong (string)                 = Longitude of Observer in degrees represented as strings
        obsElev (float)                  = Elevation of Observer in meters
        objName (string)                 = Name of the satellite
        tle1 (string)                    = First line of TLE
        tle2 (string)                    = Second line of TLE
        satRise (string or ephem.date)  = The time at which the satellite rises above horizon
        satSet (string or ephem.date)   = The time at which the satellite sets
        interval (float)                = The rate at which to sample during one rise/set cycle in seconds
        beaconFreq (float)               = The frequency of the beacon

    RETURNS:
        Param1 (float)                   = The bandwidth of the satellite during the rise/set cycle
        Param2 (list)                    = All the frequencies during the rise/set cycle sampled
                                           by given interval
    AFFECTS:
        None
    EXCEPTIONS:
        None
    DEPENDENCIES:
        ephem.date(...), ephem.hour, ephem.minute, ephem.second
    """
    currTime = satRise
    dopplerFrequencies = []
    dopplerBandwidth = []

    if opt.debug:
        print 'satellite_bandwidth ', currTime, satSet, interval

    while((currTime.triple())[2] < (satSet.triple())[2]): # the 2nd index of the returned tuple has the fraction of the day
        try:
            (sublat, sublong, range_val, range_velocity, az, alt, ra, dec, elevation) = satellite_values_at_time(opt, obsLat, obsLong, obsElev, objName, tle1, tle2, currTime)
            (dopplerFreq) = doppler_shift(beaconFreq, range_velocity)
            dopplerFrequencies.append(dopplerFreq)
            dopplerBandwidth.append(dopplerFreq - beaconFreq)
            currTime = currTime + ephem.second*interval
            currTime = ephem.date(currTime)
        except Exception as eobj:
            exp_str = str(ExceptionString(eobj))
            print "exception: %s." % (exp_str)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            print lines

    if opt.debug:
        print '# DF:', numpy.array(dopplerFrequencies) / 1e6, ' MHz'
        print '# OB:', numpy.array(dopplerBandwidth) / 1e3, ' kHz'

    return (numpy.array(dopplerBandwidth), numpy.array(dopplerFrequencies))



def __read_config__(inifile):
    """
    DESCRIPTION:
        The function parses the given file and returns a dictionary with the values.
    INPUTS:
        inifile (string) = the name of the file to be read including path

    RETURNS:
        For an object config :

        Dictionary with name given by [Section] each of which contains:
            ObsLat              Decimal observer latitude
            ObsLong             Decimal observer longitude
            ObsElev             Decimal observer elevation
            ObjName             String object name
            TLE1                String of TLE line 1 in standard format.
            TLE2                String of TLE line 2 in standard format.
            BeaconFreq          Array of floating point beacon frequencies in Hz.

        For a radio config :
            TBD

    AFFECTS:
        None
    EXCEPTIONS:
        None
    DEPENDENCIES:
        Use module re for the simple regex used in parsing the file.

    Note:

        Example object format

        [PROPCUBE_MERRY]
        ObsLat=42.623108
        ObsLong=-71.489069
        ObsElev=150.0
        ObjName="PROPCUBE_MERRY"
        TLE1="1 90735U          16074.41055570 +.00001915 +00000-0 +22522-3 0 00790"
        TLE2="2 90735 064.7823 174.3149 0209894 234.3073 123.8339 14.73463371009286"
        BeaconFreq=[380.0e6,2.38e9]
        StartDate="2016/03/21 14:00:00"
        Interval="00:00:10"

        Example radio format :

        TBD

    """

    objects = {}

    print "# loading config ", inifile
    cparse = ConfigParser.ConfigParser()
    cparse.read(inifile)

    for s in cparse.sections():
        vals = cparse.options(s)
        cfg = {}
        for v in vals:
            cfg[v] = cparse.get(s,v)
        objects[s] = cfg

    return objects

def get_next_object(opt, site, objects, ctime):
    """ Not too efficent but works for now. """
    rise_list = {}

    for obj in objects:
        obj_id = obj
        obj_info = objects[obj]

        if opt.debug:
            print "# object ", obj_id, " @ ", ctime
            print "# obj_info", obj_info

        site_name = site['site']['name']
        site_tag = site['site']['tag']
        obs_lat = site['site']['latitude']
        obs_long = site['site']['longitude']
        obs_elev = float(site['site']['elevation'])
        obj_name = obj_info['name']
        obj_tle1 = obj_info['tle1'][1:-1]
        obj_tle2 = obj_info['tle2'][1:-1]
        obj_freqs = numpy.array(string.split(obj_info['frequencies'],','),numpy.float32)
        c_dtime = datetime.datetime.utcfromtimestamp(ctime)
        c_ephem_time = ephem.Date(c_dtime)

        (sat_rise, sat_transit, sat_set) = satellite_rise_and_set(opt, obs_lat, obs_long, obs_elev, obj_name, obj_tle1, obj_tle2, c_ephem_time)

        if sat_set <= sat_rise or sat_transit <= sat_rise or sat_set <= sat_transit:
            continue

        rise_list[sat_rise] = obj

    if opt.debug:
        print ' rise list : ', rise_list

    keys = rise_list.keys()

    if opt.debug:
        print ' rise keys : ', keys

    keys.sort()

    if opt.debug:
        print ' sorted : ', keys
        print ' selected : ', rise_list[keys[0]]

    return rise_list[keys[0]]


def ephemeris_passes(opt, st0, et0):
    """
    DESCRIPTION:
        Finds passes from the start time to the end time given the options. Will
        implement a bash script or execute on the command line.
    USAGE:
        ephemeris_passes(opt, st0, et0)
    INPUTS:
        opt             command line arguments
        st0             unix time start time
        et0             unix time end time

    RETURNS:
        None
    AFFECTS:
        Prints all the passes
    EXCEPTIONS:
        None
    DEPENDENCIES:
        ephem
    """

    passes = {}

    objects = __read_config__(opt.config)

    site = __read_config__(opt.site)

    if opt.verbose:
        print "# got objects ", objects
        print "# got radio site ", site
        print "\n"

    ctime = st0
    etime = et0

    last_sat_rise = ctime

    while ctime < etime:

        obj = get_next_object(opt, site, objects, ctime)

        obj_id = obj
        obj_info = objects[obj]

        if opt.debug:
            print "# object ", obj_id, " @ ", ctime
            print "# obj_info", obj_info

        site_name = site['site']['name']
        site_tag = site['site']['tag']
        obs_lat = site['site']['latitude']
        obs_long = site['site']['longitude']
        obs_elev = float(site['site']['elevation'])
        obj_name = obj_info['name']
        obj_tle1 = obj_info['tle1'][1:-1]
        obj_tle2 = obj_info['tle2'][1:-1]
        obj_freqs = numpy.array(string.split(obj_info['frequencies'],','),numpy.float32)
        c_dtime = datetime.datetime.utcfromtimestamp(ctime)
        c_ephem_time = ephem.Date(c_dtime)

        try:
            (sat_rise, sat_transit, sat_set) = satellite_rise_and_set(opt, obs_lat, obs_long, obs_elev, obj_name, obj_tle1, obj_tle2, c_ephem_time)

            if sat_set <= sat_rise or sat_transit <= sat_rise or sat_set <= sat_transit:
                continue

            if not last_sat_rise == sat_rise:
                (sub_lat, sub_long, sat_range, sat_velocity, az, el, ra, dec, alt) = satellite_values_at_time(opt, obs_lat, obs_long, obs_elev, obj_name, obj_tle1, obj_tle2, sat_transit)
                (obj_bandwidth, obj_doppler) = satellite_bandwidth(opt, obs_lat, obs_long, obs_elev, obj_name, obj_tle1, obj_tle2, sat_rise, sat_set, op.interval, obj_freqs)
                last_sat_rise = sat_rise
                if opt.debug:
                    print "time : ", c_ephem_time, sat_set, (sat_set - c_ephem_time)*60*60*24
                ctime = ctime + (sat_set - c_ephem_time)*60*60*24

                if opt.el_mask:
                    el_val = numpy.rad2deg(el)
                    el_mask = numpy.float(opt.el_mask)

                    if opt.debug:
                        print '# el_val ', el_val, ' el_mask ', el_mask

                    if el_val < el_mask: # check mask here!
                        continue

                # This should really go out as digital metadata into the recording location

                print "# Site : %s " % (site_name)
                print "# Site tag : %s " % (site_tag)
                print "# Object Name: %s" % (obj_name)
                print "# observer @ latitude : %s, longitude : %s, elevation : %s m" % (obs_lat, obs_long, obs_elev)

                print "# GMT -- Rise Time: %s, Transit Time: %s, Set Time: %s" % (sat_rise, sat_transit, sat_set)
                print "# Azimuth: %f deg, Elevation: %f deg, Altitude: %g km" % (numpy.rad2deg(az), numpy.rad2deg(el), alt/1000.0)
                print "# Frequencies: %s MHz, Bandwidth: %s kHz" % ( obj_freqs / 1.0e6, obj_bandwidth[numpy.argmax(obj_bandwidth)] / 1.0e3)

                pass_md = {'obj_id':obj_id,
                           'rise_time':sat_rise,
                           'transit_time':sat_transit,
                           'set_time':sat_set,
                           'azimuth':numpy.rad2deg(az),
                           'elevation':numpy.rad2deg(el),
                           'altitude':alt,
                           'doppler_frequency':obj_doppler,
                           'doppler_bandwidth':obj_bandwidth}


                if opt.schedule:
                    d = sat_rise.tuple()
                    rise_time = "%04d%02d%02d_%02d%02d" % (d[0],d[1],d[2],d[3],d[4])

                    offset_rise = ephem.date(sat_rise - ephem.minute)
                    d = offset_rise.tuple()
                    offset_rise_time = "%04d-%02d-%02dT%02d:%02d:%02dZ" % (d[0],d[1],d[2],d[3],d[4],int(d[5]))

                    offset_set = ephem.date(sat_set + ephem.minute)
                    d = offset_set.tuple()
                    offset_set_time = "%04d-%02d-%02dT%02d:%02d:%02dZ" % (d[0],d[1],d[2],d[3],d[4],int(d[5]))


                    cmd_lines = []
                    radio_channel = string.split(site['radio']['channel'][1:-1],',')
                    radio_gain = string.split(site['radio']['gain'][1:-1],',')
                    radio_address = string.split(site['radio']['address'][1:-1],',')
                    recorder_channels = string.split(site['recorder']['channels'][1:-1],',')
                    radio_sample_rate = site['radio']['sample_rate']

                    cmd_line0 = "%s " % (site['recorder']['command'])

                    if site['radio']['type'] == 'b210':

                        # just record a fixed frequency, needs a dual radio Thor3 script. This can be done!
                        idx = 0
                        freq = obj_freqs[1]

                        cmd_line1 = "-r %s -d \"%s\" -s %s -e %s -c %s -f %4.3f " % (radio_sample_rate, radio_channel[idx],offset_rise_time, offset_set_time,recorder_channels[idx],freq)

                        log_file_name = "%s_%s_%s_%dMHz.log" % (site_tag,obj_id,offset_rise_time,int(freq/1.0e6))
                        cmd_fname = '%s_%s_%s_%dMHz' % (site_tag, obj_id, rise_time, int(freq/1.0e6))

                        cmd_line2 = " -g %s -m %s --devargs num_recv_frames=1024 --devargs master_clock_rate=24.0e6 -o %s/%s" % (radio_gain[idx],radio_address[idx],site['recorder']['data_path'],cmd_fname)
                        cmd_line2 += ' {0}'.format(site['radio'].get('extra_args', '')).rstrip()

                        if not opt.foreground:
                            cmd_line0 = 'nohup ' + cmd_line0
                            cmd_line2 = cmd_line2 + ' 2>&1 &'
                        else:
                            cmd_line2 = cmd_line2

                        if opt.debug:
                            print cmd_line0, cmd_line1, cmd_line2, cmd_fname

                        cmd_lines.append((cmd_line0 + cmd_line1 + cmd_line2, cmd_fname, pass_md, obj_info))

                        print "\n"

                    elif site['radio']['type'] == 'n200_tvrx2':

                        cmd_line1 = " -r %s -d \"%s %s\" -s %s -e %s -c %s,%s -f %4.3f,%4.3f " % (radio_sample_rate,radio_channel[0],radio_channel[1],offset_rise_time, offset_set_time,recorder_channels[0],recorder_channels[1],obj_freqs[0],obj_freqs[1])

                        log_file_name = "%s_%s_%s_combined.log" % (site_tag,obj_id,offset_rise_time)
                        cmd_fname = '%s_%s_%s_combined' % (site_tag, obj_id, rise_time)

                        cmd_line2 = " -g %s,%s -m %s -o %s/%s" % (radio_gain[0],radio_gain[1],radio_address[0],site['recorder']['data_path'],cmd_fname)
                        cmd_line2 += ' {0}'.format(site['radio'].get('extra_args', '')).rstrip()

                        if not opt.foreground:
                            cmd_line0 = 'nohup ' + cmd_line0
                            cmd_line2 = cmd_line2 + ' 2>&1 &'
                        else:
                            cmd_line2 = cmd_line2

                        if opt.debug:
                            print cmd_line0, cmd_line1, cmd_line2, cmd_fname

                        cmd_lines.append((cmd_line0 + cmd_line1 + cmd_line2, cmd_fname, pass_md, obj_info))

                print "\n"

                if opt.foreground:
                    dtstart0 = dateutil.parser.parse(offset_rise_time)
                    dtstop0 = dateutil.parser.parse(offset_set_time)
                    start0 = int((dtstart0 - datetime.datetime(1970,1,1,tzinfo=pytz.utc)).total_seconds())
                    stop0 = int((dtstop0 - datetime.datetime(1970,1,1,tzinfo=pytz.utc)).total_seconds())

                    if opt.verbose:
                        print "# waiting for %s @ %s " % (obj_id, offset_rise_time)

                    while time.time() < start0 - 30:
                        time.sleep(op.interval)
                        if opt.verbose:
                            print "#    %d sec" % (start0 - time.time())

                    for cmd_tuple in cmd_lines:

                        cmd, cmd_fname, pass_md, info_md = cmd_tuple

                        print "# Executing command %s " % (cmd)

                        # write the digital metadata
                        start_idx = int(start0)
                        mdata_dir = site['recorder']['metadata_path']  + '/' + cmd_fname + '/metadata'


                        # site metadata
                        # note we use directory structure for the dictionary here
                        # eventually we will add this feature to digital metadata

                        for k in site:

                            try:
                                os.makedirs(mdata_dir + '/config/%s' % (k) )
                            except:
                                pass

                            md_site_obj = DigitalMetadataWriter(mdata_dir + '/config/%s' % (k), 3600, 60, 1, 1, k)

                            if opt.debug:
                                print site[k]

                            if opt.verbose:
                                print "# writing metadata config / %s " % (k)

                            md_site_obj.write(start_idx,site[k])

                        # info metadata
                        try:
                            os.makedirs(mdata_dir + '/info')
                        except:
                            pass

                        md_info_obj = DigitalMetadataWriter(mdata_dir + '/info', 3600, 60, 1, 1, 'info')

                        if opt.verbose:
                            print "# writing metadata info"

                        if opt.debug:
                            print info_md

                        md_info_obj.write(start_idx,info_md)

                        # pass metadata
                        try:
                            os.makedirs(mdata_dir + '/pass')
                        except:
                            pass

                        md_pass_obj = DigitalMetadataWriter(mdata_dir + '/pass', 3600, 60, 1, 1, 'pass')

                        if opt.verbose:
                            print "# writing metadata pass"

                        if opt.debug:
                            print pass_md

                        md_pass_obj.write(start_idx,pass_md)

                        #sys.exit(1)

                        # call the command
                        try:
                            subprocess.call(cmd,shell=True)
                        except Exception as eobj:
                            exp_str = str(ExceptionString(eobj))
                            print "exception: %s." % (exp_str)
                            exc_type, exc_value, exc_traceback = sys.exc_info()
                            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                            print lines

                    print "# wait..."
                    while time.time() < stop0 + 1:
                        time.sleep(op.interval)
                        if opt.verbose:
                            print "# complete in %d sec" % (stop0 - time.time())


        except Exception as eobj:
            exp_str = str(ExceptionString(eobj))
            print "exception: %s." % (exp_str)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            print lines
            #sys.exit(1)
            # advance 10 minutes
            ctime = ctime + 60 * op.interval

            continue

def parse_command_line():
    parser = optparse.OptionParser()
    parser.add_option("-v", "--verbose",action="store_true",
                       dest="verbose", default=False,help="prints debug output and additional detail.")
    parser.add_option("-d", "--debug",action="store_true",
                      dest="debug", default=False,help="run in debug mode and not service context.")
    parser.add_option("-b", "--bash",action="store_true",
                      dest="schedule", default=False,help="create schedule file for bash shell based command / control.")
    parser.add_option("-m","--mask",dest="el_mask",type=float,default=0.0,help="mask all passes below the provided elevation.")
    parser.add_option("-c", "--config",dest="config",default='config/beacons.ini',help="Use configuration file <config>.")
    parser.add_option("-f", "--foreground",action="store_true",dest="foreground",help="Execute schedule in foreground.")
    parser.add_option("-s", "--starttime",dest="starttime",help="Start time in ISO8601 format, e.g. 2016-01-01T15:24:00Z")
    parser.add_option("-e", "--endtime",dest="endtime",help="End time in ISO8601 format, e.g. 2016-01-01T16:24:00Z")
    parser.add_option("-i", "--interval",dest="interval",type=float,default=10.0,help="Sampling interval for ephemeris predictions, default is 10 seconds.")
    parser.add_option("-r", "--radio",dest="site",default='config/site.ini',help="Radio site configuration file.")

    (options, args) = parser.parse_args()

    return (options, args)

if __name__ == "__main__":
    # parse command line options
    op,args = parse_command_line()

    if op.starttime is None:
        st0 = int(math.ceil(time.time())) + 10
    else:
        dtst0 = dateutil.parser.parse(op.starttime)
        st0 = int((dtst0 - datetime.datetime(1970,1,1,tzinfo=pytz.utc)).total_seconds())

        print('Start time: %s (%d)' % (dtst0.strftime('%a %b %d %H:%M:%S %Y'), st0))

    if op.endtime is None:
        # default to the next 24 hours
        et0 = st0 + 60*60*24.0
    else:
        dtet0 = dateutil.parser.parse(op.endtime)
        et0 = int((dtet0 - datetime.datetime(1970,1,1,tzinfo=pytz.utc)).total_seconds())

        print('End time: %s (%d)' % (dtet0.strftime('%a %b %d %H:%M:%S %Y'), et0))

    ephemeris_passes(op, st0, et0)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Module defining a Digital RF Source block."""
import os

import gnuradio.blocks
import h5py
import pmt
import numpy as np
import six
from gnuradio import gr

from digital_rf import DigitalRFReader, util

H5T_LOOKUP = {
    # (class, itemsize, is_complex): {name, dtype, fillvalue}
    (h5py.h5t.INTEGER, 1, False): dict(
        name='s8', dtype=np.int8, fillvalue=np.iinfo(np.int8).min,
    ),
    (h5py.h5t.INTEGER, 2, False): dict(
        name='s16', dtype=np.int16, fillvalue=np.iinfo(np.int16).min,
    ),
    (h5py.h5t.INTEGER, 4, False): dict(
        name='s32', dtype=np.int32, fillvalue=np.iinfo(np.int32).min,
    ),
    (h5py.h5t.INTEGER, 8, False): dict(
        name='s64', dtype=np.int64, fillvalue=np.iinfo(np.int64).min,
    ),
    (h5py.h5t.FLOAT, 4, False): dict(
        name='f32', dtype=np.float32, fillvalue=np.nan,
    ),
    (h5py.h5t.FLOAT, 8, False): dict(
        name='f64', dtype=np.float64, fillvalue=np.nan,
    ),
    (h5py.h5t.INTEGER, 1, True): dict(
        name='sc8', dtype=np.dtype([('r', np.int8), ('i', np.int8)]),
        fillvalue=(np.iinfo(np.int8).min,)*2,
    ),
    (h5py.h5t.INTEGER, 2, True): dict(
        name='sc16', dtype=np.dtype([('r', np.int16), ('i', np.int16)]),
        fillvalue=(np.iinfo(np.int16).min,)*2,
    ),
    (h5py.h5t.INTEGER, 4, True): dict(
        name='sc32', dtype=np.dtype([('r', np.int32), ('i', np.int32)]),
        fillvalue=(np.iinfo(np.int32).min,)*2,
    ),
    (h5py.h5t.INTEGER, 8, True): dict(
        name='sc64', dtype=np.dtype([('r', np.int64), ('i', np.int64)]),
        fillvalue=(np.iinfo(np.int64).min,)*2,
    ),
    (h5py.h5t.FLOAT, 4, True): dict(
        name='fc32', dtype=np.complex64, fillvalue=(np.nan+np.nan*1j),
    ),
    (h5py.h5t.FLOAT, 8, True): dict(
        name='fc64', dtype=np.complex128, fillvalue=(np.nan+np.nan*1j),
    ),
}


def get_h5type(cls, size, is_complex):
    try:
        typedict = H5T_LOOKUP[(cls, size, is_complex)]
    except KeyError:
        raise ValueError('HDF5 data type not supported for reading.')
    return typedict


class digital_rf_channel_source(gr.sync_block):
    """Source block for reading a channel of Digital RF data."""

    def __init__(
        self, channel_dir, start=None, end=None, repeat=False, gapless=False,
        min_chunksize=None,
    ):
        """Read a channel of data from a Digital RF directory.

        In addition to outputting samples from Digital RF format data, this
        block also emits a 'properties' message containing inherent channel
        properties and adds stream tags using the channel's accompanying
        Digital Metadata. See the Notes section for details on what the
        messages and stream tags contain.


        Parameters
        ----------

        channel_dir : string | list of strings
            Either a single channel directory containing 'drf_properties.h5'
            and timestamped subdirectories with Digital RF files, or a list of
            such. A directory can be a file system path or a url, where the url
            points to a channel directory. Each must be a local path, or start
            with 'http://'', 'file://'', or 'ftp://''.


        Other Parameters
        ----------------

        start : None | int | float | string, optional
            A value giving the start of the channel's playback.
            If None or '', the start of the channel's available data is used.
            If an integer, it is interpreted as a sample index given in the
            number of samples since the epoch (time_since_epoch*sample_rate).
            If a float, it is interpreted as a UTC timestamp (seconds since
            epoch).
            If a string, three forms are permitted:
                1) a string which can be evaluated to an integer/float and
                    interpreted as above,
                2) a string beginning with '+' and followed by an integer
                    (float) expression, interpreted as samples (seconds) from
                    the start of the data, and
                3) a time in ISO8601 format, e.g. '2016-01-01T16:24:00Z'

        end : None | int | float | string, optional
            A value giving the end of the channel's playback.
            If None or '', the end of the channel's available data is used.
            See `start` for a description of how this value is interpreted.

        repeat : bool, optional
            If True, loop the data continuously from the start after the end
            is reached. If False, stop after the data is read once.

        gapless : bool, optional
            If True, output default-filled samples for any missing data between
            start and end. If False, skip missing samples and add an `rx_time`
            stream tag to indicate the gap.

        min_chunksize : None | int, optional
            Minimum number of samples to output at once. This value can be used
            to adjust the source's performance to reduce underruns and
            processing time. If None, a sensible default will be used.


        Notes
        -----

        A channel directory must contain subdirectories/files in the format:
            [YYYY-MM-DDTHH-MM-SS]/rf@[seconds].[%03i milliseconds].h5

        Each directory provided is considered the same channel. An error is
        raised if their sample rates differ, or if their time periods overlap.

        Upon start, this block sends a 'properties' message on its output
        message port that contains a dictionary with one key, the channel's
        name, and a value which is a dictionary of properties found in the
        channel's 'drf_properties.h5' file.

        This block emits the following stream tags at the appropriate sample
        for each of the channel's accompanying Digital Metadata samples:

            rx_time : (int secs, float frac) tuple
                Time since epoch of the sample.

            rx_rate : float
                Sample rate in Hz.

            rx_freq : float | 1-D array of floats
                Center frequency or frequencies of the subchannels based on
                the 'center_frequencies' metadata field.

            metadata : dict
                Any additional Digital Metadata fields are added to this
                dictionary tag of metadata.

        """
        if isinstance(channel_dir, six.string_types):
            channel_dir = [channel_dir]
        # eventually, we should re-factor DigitalRFReader and associated so
        # that reading from a list of channel directories is possible
        # with a DigitalRFChannelReader class or similar
        # until then, split the path and use existing DigitalRFReader
        top_level_dirs = []
        chs = set()
        for ch_dir in channel_dir:
            top_level_dir, ch = os.path.split(ch_dir)
            top_level_dirs.append(top_level_dir)
            chs.add(ch)
        if len(chs) == 1:
            ch = chs.pop()
        else:
            raise ValueError('Channel directories must have the same name.')
        self._ch = ch

        self._Reader = DigitalRFReader(top_level_dirs)
        self._properties = self._Reader.get_properties(self._ch)

        typeclass = self._properties['H5Tget_class']
        itemsize = self._properties['H5Tget_size']
        is_complex = self._properties['is_complex']
        vlen = self._properties['num_subchannels']
        sr = self._properties['samples_per_second']

        self._itemsize = itemsize
        self._sample_rate = sr
        self._sample_rate_pmt = pmt.from_double(float(sr))

        # determine output signature from HDF5 type metadata
        typedict = get_h5type(typeclass, itemsize, is_complex)
        self._outtype = typedict['name']
        self._itemtype = typedict['dtype']
        self._fillvalue = typedict['fillvalue']
        if vlen == 1:
            out_sig = [self._itemtype]
        else:
            out_sig = [(self._itemtype, vlen)]

        gr.sync_block.__init__(
            self,
            name="digital_rf_channel_source",
            in_sig=None,
            out_sig=out_sig,
        )

        self.message_port_register_out(pmt.intern('properties'))
        self._id = pmt.intern(self._ch)
        self._tag_queue = {}

        self._start = start
        self._end = end
        self._repeat = repeat
        self._gapless = gapless
        if min_chunksize is None:
            # FIXME: it shouldn't have to be quite this high
            self._min_chunksize = int(sr)
        else:
            self._min_chunksize = min_chunksize

        # reduce CPU usage and underruns by setting a minimum number of samples
        # to handle at once
        # (really want to set_min_noutput_items, but no way to do that from
        #  Python)
        self.set_output_multiple(self._min_chunksize)

        try:
            self._DMDReader = self._Reader.get_digital_metadata(self._ch)
        except IOError:
            self._DMDReader = None

    def _queue_tags(self, sample, tags):
        """Queue stream tags to be attached to data in the work function.

        In addition to the tags specified in the `tags` dictionary, this will
        add `rx_time` and `rx_rate` tags giving the sample time and rate.


        Parameters
        ----------

        sample : int
            Sample index for the sample to tag, given in the number of samples
            since the epoch (time_since_epoch*sample_rate).

        tags : dict
            Dictionary containing the tags to add with keys specifying the tag
            name. The value is cast as an appropriate pmt type, while the name
            will be turned into a pmt string in the work function.

        """
        # add to current queued tags for sample if applicable
        tag_dict = self._tag_queue.get(sample, {})
        if not tag_dict:
            # add time and rate tags
            time = sample/self._sample_rate
            tag_dict['rx_time'] = pmt.make_tuple(
                pmt.from_uint64(int(np.uint64(time))),
                pmt.from_double(float(time % 1)),
            )
            tag_dict['rx_rate'] = self._sample_rate_pmt
        for k, v in tags.items():
            tag_dict[k] = pmt.to_pmt(v)
        self._tag_queue[sample] = tag_dict

    def start(self):
        self._bounds = self._Reader.get_bounds(self._ch)
        self._start_sample = util.parse_identifier_to_sample(
            self._start, self._sample_rate, self._bounds[0],
        )
        self._end_sample = util.parse_identifier_to_sample(
            self._end, self._sample_rate, self._bounds[0],
        )
        if self._start_sample is None:
            self._read_start_sample = self._bounds[0]
        else:
            self._read_start_sample = self._start_sample
        # add default tags to first sample
        self._queue_tags(self._read_start_sample, {})
        # replace longdouble samples_per_second with float for pmt conversion
        properties_message = self._properties.copy()
        properties_message['samples_per_second'] = \
            float(properties_message['samples_per_second'])
        self.message_port_pub(
            pmt.intern('properties'),
            pmt.to_pmt({self._ch: properties_message}),
        )
        return super(digital_rf_channel_source, self).start()

    def work(self, input_items, output_items):
        out = output_items[0]
        nsamples = len(out)
        next_index = 0
        # repeat reading until we succeed or return
        while next_index < nsamples:
            read_start = self._read_start_sample
            # read_end is inclusive, hence the -1
            read_end = self._read_start_sample + (nsamples - next_index) - 1
            # creating a read function that has an output argument so data can
            # be copied directly would be nice
            # also should move EOFError checking into reader once watchdog
            # bounds functionality is implemented
            try:
                if self._end_sample is None:
                    if read_end > self._bounds[1]:
                        self._bounds = self._Reader.get_bounds(self._ch)
                        read_end = min(read_end, self._bounds[1])
                else:
                    if read_end > self._end_sample:
                        read_end = self._end_sample
                if read_start > read_end:
                    raise EOFError
                # read data
                data_dict = self._Reader.read(
                    read_start, read_end, self._ch,
                )
                # handled all samples through read_end regardless of whether
                # they were written to the output vector
                self._read_start_sample = read_end + 1
                # early escape for no data
                if not data_dict:
                    if self._gapless:
                        # output empty samples if no data and gapless output
                        stop_index = next_index + read_end + 1 - read_start
                        out[next_index:stop_index] = self._fillvalue
                        next_index = stop_index
                    else:
                        # clear any existing tags
                        self._tag_queue.clear()
                        # add tag at next potential sample to indicate skip
                        self._queue_tags(self._read_start_sample, {})
                    continue
                # read corresponding metadata
                if self._DMDReader is not None:
                    meta_dict = self._DMDReader.read(
                        read_start, read_end,
                    )
                    for sample, meta in meta_dict.items():
                        # add tags from Digital Metadata
                        # (in addition to default time and rate tags)
                        # eliminate sample_rate_* tags with duplicate info
                        meta.pop('sample_rate_denominator', None)
                        meta.pop('sample_rate_numerator', None)
                        # get center frequencies for rx_freq tag, squeeze()[()]
                        # to get single value if possible else pass as an array
                        cf = meta.pop('center_frequencies', None)
                        if cf is not None:
                            cf = cf.ravel().squeeze()[()]
                        tags = dict(
                            rx_freq=cf,
                            # all other metadata goes in metadata tag
                            metadata=meta,
                        )
                        self._queue_tags(sample, tags)

                # add data and tags to output
                next_continuous_sample = read_start
                for sample, data in data_dict.items():
                    # detect data skip
                    if sample > next_continuous_sample:
                        if self._gapless:
                            # advance output by skipped number of samples
                            nskipped = sample - next_continuous_sample
                            sample_index = next_index + nskipped
                            out[next_index:sample_index] = self._fillvalue
                            next_index = sample_index
                        else:
                            # emit new time tag at sample to indicate skip
                            self._queue_tags(sample, {})
                    # output data
                    n = data.shape[0]
                    stop_index = next_index + n
                    end_sample = sample + n
                    out[next_index:stop_index] = data.squeeze()
                    # output tags
                    for tag_sample in sorted(self._tag_queue.keys()):
                        if tag_sample < sample:
                            # drop tags from before current data block
                            del self._tag_queue[tag_sample]
                            continue
                        elif tag_sample >= end_sample:
                            # wait to output tags from after current data block
                            break
                        offset = (
                            self.nitems_written(0)  # offset @ start of work
                            + next_index  # additional offset of data block
                            + (tag_sample - sample)
                        )
                        tag_dict = self._tag_queue.pop(tag_sample)
                        for name, val in tag_dict.items():
                            self.add_item_tag(
                                0, offset, pmt.intern(name), val, self._id,
                            )
                    # advance next output index and continuous sample
                    next_index = stop_index  # <=== next_index += n
                    next_continuous_sample = end_sample
            except EOFError:
                if self._repeat:
                    if self._start_sample is None:
                        self._read_start_sample = self._bounds[0]
                    else:
                        self._read_start_sample = self._start_sample
                    self._queue_tags(self._read_start_sample, {})
                    continue
                else:
                    break
        if next_index == 0:
            # return WORK_DONE
            return -1
        return next_index

    def get_gapless(self):
        return self._gapless

    def set_gapless(self, gapless):
        self._gapless = gapless

    def get_repeat(self):
        return self._repeat

    def set_repeat(self, repeat):
        self._repeat = repeat


class digital_rf_source(gr.hier_block2):
    """Source block for reading Digital RF data."""

    def __init__(
        self, top_level_dir, channels=None, start=None, end=None,
        repeat=False, throttle=False, gapless=False, min_chunksize=None,
    ):
        """Read data from a directory containing Digital RF channels.

        In addition to outputting samples from Digital RF format data, this
        block also emits a 'properties' message containing inherent channel
        properties and adds stream tags using the channel's accompanying
        Digital Metadata. See the Notes section for details on what the
        messages and stream tags contain.


        Parameters
        ----------

        top_level_dir : string
            Either a single top-level directory containing Digital RF channel
            directories, or a list of such. A directory can be a file system
            path or a url, where the url points to a top level directory. Each
            must be a local path, or start with 'http://'', 'file://'', or
            'ftp://''.


        Other Parameters
        ----------------

        channels : None | string | int | iterable of previous, optional
            If None, use all available channels in alphabetical order.
            Otherwise, use the channels in the order specified in the given
            iterable (a string or int is taken as a single-element iterable).
            A string is used to specify the channel name, while an int is used
            to specify the channel index in the sorted list of available
            channel names.

        start : None | string | int | iterable of previous, optional
            Can be a single value or an iterable of values corresponding to
            `channels` giving the start of the channel's playback.
            If None or '', the start of the channel's available data is used.
            If an integer, it is interpreted as a sample index given in the
            number of samples since the epoch (time_since_epoch*sample_rate).
            If a float, it is interpreted as a UTC timestamp (seconds since
            epoch).
            If a string, three forms are permitted:
                1) a string which can be evaluated to an integer/float and
                    interpreted as above,
                2) a string beginning with '+' and followed by an integer
                    (float) expression, interpreted as samples (seconds) from
                    the start of the data, and
                3) a time in ISO8601 format, e.g. '2016-01-01T16:24:00Z'

        end : None | string | int | iterable of previous, optional
            Can be a single value or an iterable of values corresponding to
            `channels` giving the end of the channel's playback.
            If None or '', the end of the channel's available data is used.
            See `start` for a description of how this value is interpreted.

        repeat : bool, optional
            If True, loop the data continuously from the start after the end
            is reached. If False, stop after the data is read once.

        throttle : bool, optional
            If True, playback the samples at their recorded sample rate. If
            False, read samples as quickly as possible.

        gapless : bool, optional
            If True, output zeroed samples for any missing data between start
            and end. If False, skip missing samples and add an `rx_time` stream
            tag to indicate the gap.

        min_chunksize : None | int, optional
            Minimum number of samples to output at once. This value can be used
            to adjust the source's performance to reduce underruns and
            processing time. If None, a sensible default will be used.

        Notes
        -----

        A top-level directory must contain files in the format:
            [channel]/[YYYY-MM-DDTHH-MM-SS]/rf@[seconds].[%03i milliseconds].h5

        If more than one top level directory contains the same channel_name
        subdirectory, this is considered the same channel. An error is raised
        if their sample rates differ, or if their time periods overlap.

        Upon start, this block sends 'properties' messages on its output
        message port that contains a dictionaries with one key, the channel's
        name, and a value which is a dictionary of properties found in the
        channel's 'drf_properties.h5' file.

        This block emits the following stream tags at the appropriate sample
        for each of the channel's accompanying Digital Metadata samples:

            rx_time : (int secs, float frac) tuple
                Time since epoch of the sample.

            rx_rate : float
                Sample rate in Hz.

            rx_freq : float | 1-D array of floats
                Center frequency or frequencies of the subchannels based on
                the 'center_frequencies' metadata field.

            metadata : dict
                Any additional Digital Metadata fields are added to this
                dictionary tag of metadata.

        """
        options = locals()
        del options['self']
        del options['top_level_dir']
        del options['channels']
        del options['start']
        del options['end']
        del options['throttle']

        Reader = DigitalRFReader(top_level_dir)
        available_channel_names = Reader.get_channels()
        self._channel_names = self._get_channel_names(
            channels, available_channel_names,
        )

        if start is None or isinstance(start, six.string_types):
            start = [start]*len(self._channel_names)
        try:
            s_iter = iter(start)
        except TypeError:
            s_iter = iter([start])
        if end is None or isinstance(end, six.string_types):
            end = [end]*len(self._channel_names)
        try:
            e_iter = iter(end)
        except TypeError:
            e_iter = iter([end])

        # make sources for each channel
        self._channels = []
        for ch, s, e in zip(self._channel_names, s_iter, e_iter):
            chsrc = digital_rf_channel_source(
                os.path.join(top_level_dir, ch), start=s, end=e, **options
            )
            self._channels.append(chsrc)

        out_sig_dtypes = [src.out_sig()[0] for src in self._channels]
        out_sig = gr.io_signaturev(
            len(out_sig_dtypes), len(out_sig_dtypes),
            [s.itemsize for s in out_sig_dtypes],
        )
        in_sig = gr.io_signature(0, 0, 0)

        gr.hier_block2.__init__(
            self,
            name="digital_rf_source",
            input_signature=in_sig,
            output_signature=out_sig,
        )

        msg_port_name = pmt.intern('properties')
        self.message_port_register_hier_out('properties')

        for k, src in enumerate(self._channels):
            if throttle:
                throt = gnuradio.blocks.throttle(
                    src.out_sig()[0].itemsize, float(src._sample_rate),
                    ignore_tags=True,
                )
                self.connect(src, throt, (self, k))
            else:
                self.connect(src, (self, k))
            self.msg_connect(src, msg_port_name, self, msg_port_name)

    @staticmethod
    def _get_channel_names(channels, available_channel_names):
        # channels can be None, in which case we use all available
        if channels is None:
            return available_channel_names
        # or channels can be a string for a single channel
        if isinstance(channels, six.string_types):
            channels = [channels]
        unselected_channels = available_channel_names[:]
        channel_names = []
        # now channels should be an iterable of strings or indexes
        try:
            ch_iter = iter(channels)
        except TypeError:
            # unless channels is potentially a single index
            ch_iter = iter([channels])
        for ch in ch_iter:
            # make None and index ch into string channel name
            if ch is None or ch == '':
                # use first available channel (alphabetical)
                try:
                    ch_name = unselected_channels[0]
                except IndexError:
                    raise ValueError(
                        '"None" invalid for channel, all available '
                        'channels have been selected.'
                    )
            else:
                # try ch as a list index into available channels
                try:
                    ch_name = available_channel_names[int(ch)]
                except (TypeError, ValueError):
                    # not an index, that's fine
                    ch_name = ch
                except IndexError:
                    raise IndexError(
                        'Channel index {0} does not exist.'.format(ch)
                    )

            # now assume ch is a string, get from unselected channel list
            try:
                unselected_channels.remove(ch_name)
            except ValueError:
                errstr = (
                    'Channel {0} does not exist or has already been '
                    'selected.'
                )
                raise ValueError(errstr.format(ch_name))
            channel_names.append(ch_name)
        return channel_names

    def get_gapless(self):
        return self._channels[0]._gapless

    def set_gapless(self, gapless):
        for ch in self._channels:
            ch.set_gapless(gapless)

    def get_repeat(self):
        return self._channels[0]._repeat

    def set_repeat(self, repeat):
        for ch in self._channels:
            ch.set_repeat(repeat)



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""digital_rf_archive.py is a tool for archiving Digital RF data.

$Id$
"""
# standard python imports
import datetime
import time
import glob
import argparse
import os
import sys
import socket
import urlparse
import subprocess
import shutil
import multiprocessing
import traceback


def archive_subdirectory_local_local(args):
    """archive_subdirectory_local_local is the method called by a pool of multiprocessing process to archive
    one subdirectory from a local machine to a local machine.

    Inputs: args, a list of:
        source - path to high level source directory
        channel - channel name
        subdir - basename of subdirectory to archive
        dt - datetime of subdirectory name
        nextDT - datetime of next subdirectory.  None if no next subdirectory.
        is_metadata - True if metadata, False if RF data
        channel_dest - path to where channel will be backed up
        startDT - start datetime of archive
        endDT - end datetime of archive
        gzip - gzip (0-9) 0 is no compression
        verbose - True for verbose output, False otherwise

    Returns: the number of files backed up
    """
    # constants
    _rf_file_glob = 'rf@[0-9]*.[0-9][0-9][0-9].h5'
    _metadata_file_glob = '*@[0-9]*.h5'

    file_count = 0

    source, channel, subdir, dt, nextDT, is_metadata, channel_dest, startDT, endDT, gzip, verbose = args

    # determine if this channel is compressed
    if is_metadata:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _metadata_file_glob))
    else:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _rf_file_glob))
    if len(archive_files) == 0:
        if verbose:
            print('skipping because no files found in %s' % (subdir))
        return(0)

    # decide if this is a full copy, partial copy, or skip
    if dt > endDT:
        if verbose:
            print('skipping %s because after period %s - %s' % (subdir, str(startDT),
                                                                str(endDT)))
        return(0)

    if dt < startDT and nextDT != None:
        if nextDT < startDT:
            if verbose:
                print('skipping %s because entirely before period %s - %s' % (subdir, str(startDT),
                                                                              str(endDT)))
            return(0)

    if gzip > 0:
        is_compressed = _is_h5_file_compressed(archive_files[0])
    else:
        is_compressed = False

    # the only options left now are complete or partial copies
    # check if complete copy needed
    if startDT < dt and nextDT != None:
        if nextDT < endDT:
            if (is_compressed or gzip == 0) or is_metadata:
                if verbose:
                    print('%s - complete copy' % (subdir))
                shutil.copytree(os.path.join(source, channel, subdir),
                                os.path.join(channel_dest, subdir))
                file_count += len(glob.glob(os.path.join(source,
                                                         channel, subdir, '*')))
                return(file_count)
    if verbose:
        print('%s partial copy or copy with added compression' % (subdir))
    target_dir = os.path.join(channel_dest, subdir)
    os.mkdir(target_dir)
    if is_metadata:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _metadata_file_glob))
    else:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _rf_file_glob))
    archive_files.sort()  # allows us to stop checking when a file moves out of time range
    for archive_file in archive_files:
        basename = os.path.basename(archive_file)
        ts = datetime.datetime.utcfromtimestamp(
            float(basename[basename.find('@') + 1:-3]))
        if ts >= startDT and ts <= endDT:
            # check if we need to compress it
            if not is_compressed and gzip > 0 and not is_metadata:
                src_file = os.path.join(target_dir, basename)
                cmd = 'h5repack -i %s -o %s -f rf_data:GZIP=%i' % (
                    archive_file, src_file, gzip)
                try:
                    subprocess.check_call(cmd.split())
                    file_count += 1
                except:
                    traceback.print_exc()
                    print('cmd <%s> failed' % (cmd))
            else:
                shutil.copyfile(archive_file, os.path.join(
                    target_dir, os.path.basename(archive_file)))
                file_count += 1
        elif ts > endDT:
            return(file_count)

    return(file_count)


def archive_subdirectory_local_remote(args):
    """archive_subdirectory_local_remote is the method called by a pool of multiprocessing process to archive
    one subdirectory from a local machine to a remote machine.

    Inputs: args, a list of:
        source - path to high level source directory
        channel - channel name
        subdir - basename of subdirectory to archive
        dt - datetime of subdirectory name
        nextDT - datetime of next subdirectory.  None if no next subdirectory.
        is_metadata - True if metadata, False if RF data
        channel_dest - path to where channel will be backed up
        startDT - start datetime of archive
        endDT - end datetime of archive
        gzip - gzip (0-9) ) is no compression
        verbose - True for verbose output, False otherwise

    Returns: the number of files backed up
    """
    # constants
    _rf_file_glob = 'rf@[0-9]*.[0-9][0-9][0-9].h5'
    _metadata_file_glob = '*@[0-9]*.h5'

    file_count = 0

    source, channel, subdir, dt, nextDT, is_metadata, channel_dest, startDT, endDT, gzip, verbose = args

    # determine if this channel is compressed
    if is_metadata:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _metadata_file_glob))
    else:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _rf_file_glob))
    if len(archive_files) == 0:
        if verbose:
            print('skipping because no files found in %s' % (subdir))
        return(0)

    # decide if this is a full copy, partial copy, or skip
    if dt > endDT:
        if verbose:
            print('skipping %s because after period %s - %s' % (subdir, str(startDT),
                                                                str(endDT)))
        return(0)

    if dt < startDT and nextDT != None:
        if nextDT < startDT:
            if verbose:
                print('skipping %s because entirely before period %s - %s' % (subdir, str(startDT),
                                                                              str(endDT)))
            return(0)

    if gzip > 0:
        is_compressed = _is_h5_file_compressed(archive_files[0])
    else:
        is_compressed = False

    # the only options left now are complete or partial copies
    # check if complete copy needed
    if startDT < dt and nextDT != None:
        if nextDT < endDT:
            if (is_compressed or gzip == 0) or is_metadata:
                if verbose:
                    print('%s - complete copy' % (subdir))
                cmd = 'scp -q -r %s %s' % (os.path.join(source, channel, subdir),
                                           os.path.join(channel_dest, subdir))
                try:
                    subprocess.check_call(cmd.split())
                    file_count += len(glob.glob(os.path.join(source,
                                                             channel, subdir, '*')))
                except subprocess.CalledProcessError:
                    raise IOError, 'Failed to copy entire subdirectory with cmd <%s>' % (
                        cmd)
                return(file_count)

    if verbose:
        print('%s partial copy or copy with added compression' % (subdir))
    target_dir = os.path.join(channel_dest, subdir)
    # make channel dir locally, then copy
    subdir_local = os.path.join('/tmp', subdir)
    try:
        os.system('rm -rf %s' % (subdir_local))
    except:
        pass
    os.mkdir(subdir_local)
    cmd = 'scp -q -r %s %s' % (subdir_local, channel_dest)
    try:
        subprocess.check_call(cmd.split())
    except subprocess.CalledProcessError:
        raise IOError, 'Failed to create remote subdirectory dir with cmd <%s>' % (
            cmd)
    try:
        os.system('rm -rf %s' % (subdir_local))
    except:
        pass

    if is_metadata:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _metadata_file_glob))
    else:
        archive_files = glob.glob(os.path.join(
            source, channel, subdir, _rf_file_glob))
    archive_files.sort()  # allows us to stop checking when a file moves out of time range
    for archive_file in archive_files:
        basename = os.path.basename(archive_file)
        ts = datetime.datetime.utcfromtimestamp(
            float(basename[basename.find('@') + 1:-3]))
        if ts >= startDT and ts <= endDT:
            # if we need to compress it, make a copy first
            if not is_compressed and gzip > 0 and not is_metadata:
                src_file = os.path.join('/tmp', basename)
                try:
                    os.remove(src_file)
                except:
                    pass
                cmd = 'h5repack -i %s -o %s -f rf_data:GZIP=%i' % (
                    archive_file, src_file, gzip)
                try:
                    subprocess.check_call(cmd.split())
                except:
                    traceback.print_exc()
                    print('cmd <%s> failed' % (cmd))
                    continue
            else:
                src_file = archive_file
            cmd = 'scp -q %s %s' % (src_file, target_dir)
            try:
                subprocess.check_output(cmd.split())
                file_count += 0
            except subprocess.CalledProcessError:
                raise IOError, 'Failed to copy archive_file with cmd <%s>' % (
                    cmd)
            if src_file != archive_file:
                os.remove(src_file)
        elif ts > endDT:
            break

    return(file_count)


def _is_h5_file_compressed(filename):
    """_is_h5_file_compressed returns True if Hdf5 file gzip compresses, False otherwise
    """
    cmd = 'h5stat %s' % (filename)
    output = subprocess.check_output(cmd.split())
    for line in output.split('\n'):
        if line.find('GZIP') != -1:
            items = line.split()
            level = int(items[-1])
            if level > 0 and level < 10:
                return(True)
    return(False)


class archive:
    """archive is a class to archive a digital rf data set
    """

    def __init__(self, startDT, endDT, source, dest, pool=1, channels=None, is_metadata=False,
                 gzip=1, verbose=False, check_only=False):
        """
        __init__ will create a Digital RF archive

        Inputs:
            startDT - start datetime for the data archive
            endDT end datetime for the data archive
            source - string defining data source. Either a path, or a url.
                Empty string is defined as $PWD
            dest - where archive is to be created.  May be a path, or a scp
                type path (user@host:/path) If scp-type, keypair must allow
                scp without passwords
            pool - number of multiprocessing processes to use. Default is 1
            channels - a list of channels to archive.  If None (the default),
                archive all channels
            is_metadata - False if standard rf data, True if metadata with modified file
                naming convention.  That convention is *@<timestamp>*h5.  Metadata is stored
                under metadata subdirectory, rf_data under rf_data subdirectory
            gzip - integer specifying level of gzip compression to apply if file not already compressed.
                Default is 1.  0 for no compression.
            verbose - if True, print one line for each subdirectory.  If False (the default),
                no output except for errors.
            check_only - if True, simply check if enough space. Raises error if not local source and local dest.

        Attributes:
            self._source_type - local if local input data, remote if remote
            self._chan_dist - dictionary with keys = channel found, values = tuple of
                1: ordered list of tuples of
                    1. paths or urls to subdirectories,
                    2. datetime of that subdirectory name. and
                2: list of metadata files in channel directory
        """
        # constants
        self._sub_directory_glob = '[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]-[0-9][0-9]-[0-9][0-9]'
        self._rf_file_glob = 'rf@[0-9]*.[0-9][0-9][0-9].h5'
        self._properties_glob = '*_properties.h5'
        self._metadata_file_glob = '*@[0-9]*.h5'
        self._metadata_dir = 'metadata'
        self._rf_dir = 'rf_data'

        # verify arguments
        if endDT <= startDT:
            print('endDT <%s> must be after startDT <%s>' %
                  (str(endDT), str(startDT)))
            sys.exit(-1)

        # save args
        self.startDT = startDT
        self.endDT = endDT
        self.source = source
        self.dest = dest
        self.pool_count = pool
        self.channels = channels
        self.is_metadata = bool(is_metadata)
        self.gzip = int(gzip)
        self.verbose = bool(verbose)
        self.check_only = bool(check_only)

        self._source_type = 'local'
        if len(self.source) > 5:
            if self.source[0:5] == 'http:':
                self._source_type = 'remote'

        if self._source_type == 'local':
            # make sure source is full path
            if self.source[0] != '/':
                self.source = os.path.join(os.getcwd(), self.source)

        self._dest_type = 'local'
        hyphen = self.dest.find(':')
        slash = self.dest.find('/')
        if hyphen != -1:
            if slash == -1:
                self._dest_type = 'remote'
            elif hyphen < slash:
                self._dest_type = 'remote'

        if self.check_only:
            if self._dest_type != 'local' or self._source_type != 'local':
                raise ValueError, 'check_only flag only valid for local source and destination'

        if self.verbose:
            # set up a timer
            t = time.time()

        if self._source_type == 'local':
            self._chan_dict = self.get_chan_dict_local()
        else:
            self._chan_dict = self.get_chan_dict_remote()

        # verify all channels exist if specified
        if self.channels:
            for channel in self.channels:
                if not self._chan_dict.has_key(channel):
                    raise IOError, 'Requested channel %s not found' % (channel)
        else:
            self.channels = self._chan_dict.keys()  # archive all channels

        self._hostname = self._get_hostname()
        self._top_level_dest_dir = self.get_top_level_dest_dir()

        # set _next_level_dir to rf_data or metadata
        if self.is_metadata:
            self._next_level_dir = self._metadata_dir
        else:
            self._next_level_dir = self._rf_dir

        if self._dest_type == 'local' and self._source_type == 'local':
            # verify enough disk space
            needed_GB, avail_GB = self.estimate_disk_space()
            if self.verbose or self.check_only:
                print('Estimated GB to use = %f GB, available = %f GB' %
                      (needed_GB, avail_GB))
                if self.check_only:
                    sys.exit(-1)
            if needed_GB > avail_GB:
                raise IOError, "aborting because needed GB %f greater than available GB %f" % (
                    needed_GB, avail_GB)

        self._create_top_level_dest_dirs()

        # create a pool of workers
        self._pool = multiprocessing.Pool(pool)

        for channel in self.channels:
            if self.verbose:
                print('archiving channel %s' % (channel))
            result = self._archive_channel(channel)

        if self.verbose:
            print('digital_rf_archive took %f seconds' % (time.time() - t))

    def _archive_channel(self, channel):
        """_archive_channel will archive the channel.  Returns True if any data archived, False otherwise
        """
        # to simplify the code, break into for subroutines depending on
        # self._source_type self._dest_type
        if self._source_type == 'local':
            if self._dest_type == 'local':
                return(self._archive_channel_local_local(channel))
            else:
                return(self._archive_channel_local_remote(channel))

        else:
            raise ValueError, 'nyi'

    def _archive_channel_local_local(self, channel):
        """_archive_channel_local_local archives local input data to a local destination
        """
        is_compressed = None  # not yet known
        subdir_list = self._chan_dict[channel]
        # make channel dir in self._top_level_dest_dir/self._next_level_dir
        channel_dest = os.path.join(self.dest, self._top_level_dest_dir, self._next_level_dir,
                                    channel)
        os.mkdir(channel_dest)

        # first copy all the properties
        for properties_file in subdir_list[1]:
            if self.verbose:
                print('archiving properties file %s' % (properties_file))
            shutil.copyfile(os.path.join(self.source, channel, properties_file),
                            os.path.join(channel_dest, os.path.basename(properties_file)))

        # build iterable of arguments
        args_list = []
        for i, items in enumerate(subdir_list[0]):
            subdir, dt = items
            if i < len(subdir_list[0]) - 1:
                nextDT = subdir_list[0][i + 1][1]
            else:
                nextDT = None
            # distribute the subdirectories to the pool
            args_list.append([self.source, channel, subdir, dt, nextDT, self.is_metadata, channel_dest,
                              self.startDT, self.endDT, self.gzip, self.verbose])

        file_count_list = self._pool.map(
            archive_subdirectory_local_local, args_list)

        if self.verbose:
            print('%i files backed up in channel %s' %
                  (sum(file_count_list), channel))
        if sum(file_count_list) == 0:
            print("WARNING: No files backed up in channel %s" % (channel))

    def _archive_channel_local_remote(self, channel):
        """_archive_channel_local_remote archives local input data to a remote destination
        """
        is_compressed = None  # not yet known
        subdir_list = self._chan_dict[channel]
        file_count = 0

        # make channel dir in self._top_level_dest_dir
        top_level_dest = os.path.join(
            self.dest, self._top_level_dest_dir, self._next_level_dir)
        channel_dest = os.path.join(
            self.dest, self._top_level_dest_dir, self._next_level_dir, channel)
        # make channel dir locally, then copy
        channel_local = os.path.join('/tmp', channel)
        try:
            os.system('rm -rf %s' % (channel_local))
        except:
            pass
        os.mkdir(channel_local)
        cmd = 'scp -q -r %s %s' % (channel_local, top_level_dest)
        try:
            subprocess.check_call(cmd.split())
        except subprocess.CalledProcessError:
            raise IOError, 'Failed to create remote channel dir with cmd <%s>' % (
                cmd)
        try:
            os.system('rm -rf %s' % (channel_local))
        except:
            pass

        # first copy all the properties
        for properties_file in subdir_list[1]:
            if self.verbose:
                print('archiving properties file %s' % (properties_file))
            cmd = 'scp -q %s %s' % (os.path.join(self.source,
                                                 channel, properties_file), channel_dest)
            try:
                subprocess.check_call(cmd.split())
            except subprocess.CalledProcessError:
                raise IOError, 'Failed to copy properties file with cmd <%s>' % (
                    cmd)

        # build iterable of arguments
        args_list = []
        for i, items in enumerate(subdir_list[0]):
            subdir, dt = items
            if i < len(subdir_list[0]) - 1:
                nextDT = subdir_list[0][i + 1][1]
            else:
                nextDT = None
            # distribute the subdirectories to the pool
            args_list.append([self.source, channel, subdir, dt, nextDT, self.is_metadata, channel_dest,
                              self.startDT, self.endDT, self.gzip, self.verbose])

        file_count_list = self._pool.map(
            archive_subdirectory_local_remote, args_list)

        if self.verbose:
            print('%i files backed up in channel %s' %
                  (sum(file_count_list), channel))
        if sum(file_count_list) == 0:
            print("WARNING: No files backed up in channel %s" % (channel))

    def _get_hostname(self):
        """_get_hostname returns the hostname of the local system or the remote system of the source data
        """
        if self._source_type == 'local':
            return(socket.gethostname())
        else:
            o = urlparse.urlparse(self.source)
            return(o.netloc)

    def _create_top_level_dest_dirs(self):
        """_create_top_level_dest_dirs will create the dest top level directory if needed. Raise IOError
        if it can't be created. Also creates subdirectories <metadata> and <rf_data>
        under that directory.
        """
        full_path = os.path.join(self.dest, self._top_level_dest_dir)
        if self._dest_type == 'local':
            if not os.path.isdir(self.dest):
                raise IOError, '%s does not exist or is not a directory' % (
                    self.dest)
            if not os.access(full_path, os.W_OK):
                os.mkdir(full_path)
                os.mkdir(os.path.join(full_path, self._rf_dir))
                os.mkdir(os.path.join(full_path, self._metadata_dir))
        else:
            # try to create the three remote directories - even if they already
            # exist, should succeed
            for i, this_dir in enumerate((self._top_level_dest_dir, self._metadata_dir, self._rf_dir)):
                localDir = os.path.join('/tmp', this_dir)
                try:
                    os.system('rm -rf %s' % (localDir))
                except:
                    pass
                os.mkdir(localDir)
                # now scp better succeed
                if i == 0:
                    cmd = 'scp -q -r %s %s' % (localDir, self.dest)
                else:
                    cmd = 'scp -q -r %s %s' % (localDir, os.path.join(
                        self.dest, self._top_level_dest_dir))
                try:
                    subprocess.check_call(cmd.split())
                except subprocess.CalledProcessError:
                    raise IOError, 'Failed to create remote top level dir with cmd <%s>' % (
                        cmd)
                try:
                    os.system('rm -rf %s' % (localDir))
                except:
                    pass

    def _is_h5_file_compressed(self, filename):
        cmd = 'h5stat %s' % (filename)
        output = subprocess.check_output(cmd.split())
        for line in output.split('\n'):
            if line.find('GZIP') != -1:
                items = line.split()
                level = int(items[-1])
                if level > 0 and level < 10:
                    return(True)
        return(False)

    def estimate_disk_space(self):
        """estimate_disk_space returns a tuple of (needed_GB, avail_GB) as floats for local dest"""
        # constants
        _rf_file_glob = 'rf@[0-9]*.[0-9][0-9][0-9].h5'
        _metadata_file_glob = '*@[0-9]*.h5'

        # estimate available disk space
        result = os.statvfs(self.dest)
        avail_GB = (result.f_bavail * result.f_frsize) / (1024 * 1024 * 1024.0)

        # estimate needed space
        needed_GB = 0.0
        compress = None  # will be estimated with test file
        for channel in self.channels:
            subdir_list = self._chan_dict[channel]

            # properties
            for properties_file in subdir_list[1]:
                needed_GB += os.path.getsize(os.path.join(
                    self.source, channel, properties_file))

            # rf data
            for i, items in enumerate(subdir_list[0]):
                subdir, dt = items
                if i < len(subdir_list[0]) - 1:
                    nextDT = subdir_list[0][i + 1][1]
                else:
                    nextDT = None

                if dt > self.endDT:
                    break
                if nextDT != None:
                    if nextDT <= self.startDT:
                        # too early
                        continue

                if self.is_metadata:
                    archive_files = glob.glob(os.path.join(
                        self.source, channel, subdir, _metadata_file_glob))
                    compress = 1.0
                else:
                    archive_files = glob.glob(os.path.join(
                        self.source, channel, subdir, _rf_file_glob))
                    if compress == None and len(archive_files) > 0:
                        compress = self._test_compression(archive_files[0])

                for archive_file in archive_files:
                    needed_GB += compress * os.path.getsize(archive_file)

        return((needed_GB / (1024 * 1024 * 1024.0), avail_GB))

    def _test_compression(self, testHdf5File):
        """_test_compression returns the gzip compression ratio for test file testHdf5File
        """
        if self.gzip == 0:
            return(1.0)
        resultFile = '/tmp/test.h5'
        try:
            os.remove(resultFile)
        except:
            pass
        cmd = 'h5repack -i %s -o %s -f rf_data:GZIP=%i' % (
            testHdf5File, resultFile, self.gzip)
        subprocess.check_call(cmd.split())
        result = float(os.path.getsize(resultFile)) / \
            os.path.getsize(testHdf5File)
        os.remove(resultFile)
        return(result)

    def get_top_level_dest_dir(self):
        """returns the top level dest directory to write to in format
        <hostname>_YYYY-MM-DDTHH:MM:SS_YYYY-MM-DDTHH:MM:SS
        """
        startDTStr = self.startDT.strftime('%Y%m%dT%H%M%S')
        endDTStr = self.endDT.strftime('%Y%m%dT%H%M%S')
        return('%s_%s_%s' % (self._hostname, startDTStr, endDTStr))

    def get_chan_dict_local(self):
        """get_chan_dict_local returns a dictionary with keys = channels found,
            values = list of two lists - 1: list of
            subdirectory basenames, 2: list of properties files in channel directory
        """
        ret_dict = {}
        subdirectories = glob.glob(os.path.join(
            self.source, '*', self._sub_directory_glob))
        subdirectories.sort()
        for subdirectory in subdirectories:
            basename = os.path.basename(subdirectory)
            dt = datetime.datetime.strptime(basename, '%Y-%m-%dT%H-%M-%S')
            channel_path = os.path.dirname(subdirectory)
            channel = os.path.basename(channel_path)
            if not ret_dict.has_key(channel):
                ret_dict[channel] = [[], []]
            ret_dict[channel][0].append((basename, dt))

        # now get properties files
        properties_files = glob.glob(os.path.join(
            self.source, '*', self._properties_glob))
        for properties_file in properties_files:
            basename = os.path.basename(properties_file)
            channel_path = os.path.dirname(properties_file)
            channel = os.path.basename(channel_path)
            if ret_dict.has_key(channel):
                ret_dict[channel][1].append(basename)

        return(ret_dict)

    def get_chan_dict_remote(self):
        """get_chan_dict_local returns a dictionary with keys = channels found,
            values = list of two lists - 1: list of
            subdirectory basenames, 2: list of metadata files in channel directory
        """
        return('nyi')


### main begins here ###
if __name__ == '__main__':

    # command line interface
    parser = argparse.ArgumentParser(
        description='digital_rf_archive.py is a tool for archiving Digital RF data..')
    parser.add_argument('--startDT', metavar='start datetime string',
                        help='Start UT datetime for archiving in format YYYY-MM-DDTHH:MM:SS', required=True)
    parser.add_argument('--endDT', metavar='end datetime string',
                        help='End UT datetime for archiving in format YYYY-MM-DDTHH:MM:SS', required=True)
    parser.add_argument('--source', metavar='Top level directory or url',
                        help='Top level directory or url containing the Digital RF channels or metadata channels to archive',
                        required=True)
    parser.add_argument('--dest', metavar='Full path to destination',
                        help='Full path to destination directory.  May be local, or remote in scp form (user@host:)',
                        required=True)
    parser.add_argument('--pool', metavar='Pool multiprocessing count.',
                        help='Number of multiprocesses processes to use in the pool.  Default=1.',
                        required=False, type=int, default=1)
    parser.add_argument('--channel', metavar='Optional channel name', action='append',
                        help='Optional channel name to backup.  May be given more than once.  Default is all channels.')
    parser.add_argument('--is_metadata', action='store_true', default=False,
                        help='Set this flag if metadata, as opposed to rf_data, being archived.')
    parser.add_argument('--gzip', metavar='GZIP compression 1-9.',
                        help='Level of GZIP compression 1-9.  Default=1. 0 for no compression',
                        required=False, type=int, default=1)
    parser.add_argument('-v',  '--verbose', action='store_true', default=False,
                        help='Get one line of output for each subdirectory archived.')
    parser.add_argument('--check_only', action='store_true', default=False,
                        help='Set this flag to simply check if enough space. Only possible for local source and local dest.')
    args = parser.parse_args()

    # create datetimes
    try:
        startDT = datetime.datetime.strptime(args.startDT, '%Y-%m-%dT%H:%M:%S')
    except:
        print('startDT <%s> not in expected YYYY-MM-DDTHH:MM:SS format' %
              (args.startDT))
        sys.exit(-1)
    try:
        endDT = datetime.datetime.strptime(args.endDT, '%Y-%m-%dT%H:%M:%S')
    except:
        print('endDT <%s> not in expected YYYY-MM-DDTHH:MM:SS format' %
              (args.endDT))
        sys.exit(-1)

    # verify gzip in range
    if args.gzip < 0 or args.gzip > 9:
        raise ValueError, 'gzip must be 0 (no compression), or 1-9, not %i' % (gzip)

    # call main class
    archive(startDT, endDT, args.source, args.dest, args.pool, args.channel, args.is_metadata,
            args.gzip, args.verbose, args.check_only)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""example_digital_rf_hdf5.py is an example script using the digital_rf_hdf5 module

Assumes one of the example Digital RF scripts has already been run (C: example_rf_write_hdf5, or
Python: example_digital_rf.py)

$Id$
"""
# Millstone imports
import digital_rf


testReadObj = digital_rf.DigitalRFReader(['/tmp/hdf5'])
channels = testReadObj.get_channels()
if len(channels) == 0:
    raise IOError, """Please run one of the example write scripts
        C: example_rf_write_hdf5, or Python: example_digital_rf_hdf5.py
        before running this example"""
print('found channels: %s' % (str(channels)))

print('working on channel junk0')
start_index, end_index = testReadObj.get_bounds('junk0')
print('get_bounds returned %i - %i' % (start_index, end_index))
cont_data_arr = testReadObj.get_continuous_blocks(
    start_index, end_index, 'junk0')
print('The following is a OrderedDict of all continuous block of data in (start_sample, length) format: %s' %
      (str(cont_data_arr)))

# read data - the first 3 reads of four should succeed, the fourth read
# will be beyond the available data
start_sample = cont_data_arr.keys()[0]
for i in range(4):
    try:
        result = testReadObj.read_vector(start_sample, 200, 'junk0')
        print('read number %i got %i samples starting at sample %i' %
              (i, len(result), start_sample))
        start_sample += 200
    except IOError:
        print('Read number %i went beyond existing data and raised an IOError' % (i))

# finally, get all the built in rf properties
rf_dict = testReadObj.get_properties('junk0')
print('Here is the metadata built into drf_properties.h5 (valid for all data): %s' %
      (str(rf_dict)))
